#!/usr/bin/env python3
#
# ############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################
import argparse
import glob
import json
import logging
import os.path
import pickle
import shutil
import sys
import time
import gzip
from collections import namedtuple
from io import StringIO
from traceback import print_exc
from concurrent.futures import ProcessPoolExecutor
import concurrent.futures

import pysam
import gffutils
import pyfaidx

from src.error_codes import IsoQuantExitCode
from src.modes import IsoQuantMode, ISOQUANT_MODES
from src.gtf2db import convert_gtf_to_db
from src.read_mapper import (
    DATA_TYPE_ALIASES,
    SUPPORTED_STRANDEDNESS,
    SUPPORTED_ALIGNERS,
    ASSEMBLY,
    PACBIO_CCS_DATA,
    NANOPORE_DATA,
    DataSetReadMapper
)
from src.alignment_processor import PolyATrimmed
from src.dataset_processor import DatasetProcessor, PolyAUsageStrategies
from src.graph_based_model_construction import StrandnessReportingLevel
from src.long_read_assigner import AmbiguityResolvingMethod
from src.long_read_counter import COUNTING_STRATEGIES, CountingStrategy, GroupedOutputFormat, NormalizationMethod
from src.input_data_storage import InputDataStorage, InputDataType
from src.multimap_resolver import MultimapResolvingStrategy
from src.stats import combine_counts
from src.barcode_calling import process_single_thread, process_in_parallel, get_umi_length
from src.common import setup_worker_logging, _get_log_params


logger = logging.getLogger('IsoQuant')

# Large output file types for --large_output option
LARGE_OUTPUT_TYPES = ["read_assignments", "corrected_bed", "read2transcripts", "allinfo", "none"]


def bool_str(s):
    s = s.lower()
    if s not in {'false', 'true', '0', '1'}:
        raise ValueError('Not a valid boolean string')
    return s == 'true' or s == '1'


def parse_args(cmd_args=None, namespace=None):
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    ref_args_group = parser.add_argument_group('Reference data')
    input_args_group = parser.add_argument_group('Input data')
    output_args_group = parser.add_argument_group('Output naming')
    pipeline_args_group = parser.add_argument_group('Pipeline options')
    algo_args_group = parser.add_argument_group('Algorithm settings')
    output_setup_args_group = parser.add_argument_group('Output configuration')
    align_args_group = parser.add_argument_group('Aligner settings')
    filer_args_group = parser.add_argument_group('Read filtering options')
    sc_args_group = parser.add_argument_group('Single-cell/spatial-related options:')

    other_options = parser.add_argument_group("Additional options:")
    show_full_help = '--full_help' in cmd_args

    def add_additional_option(*args, **kwargs):  # show command only with --full-help
        if not show_full_help:
            kwargs['help'] = argparse.SUPPRESS
        other_options.add_argument(*args, **kwargs)

    def add_option_to_group(opt_group, *args, **kwargs):  # show command only with --full-help
        opt_group.add_argument(*args, **kwargs)

    def add_additional_option_to_group(opt_group, *args, **kwargs):  # show command only with --full-help
        if not show_full_help:
            kwargs['help'] = argparse.SUPPRESS
        opt_group.add_argument(*args, **kwargs)

    def add_hidden_option(*args, **kwargs):  # show command only with --full-help
        kwargs['help'] = argparse.SUPPRESS
        parser.add_argument(*args, **kwargs)

    parser.add_argument("--full_help", action='help', help="show full list of options")

    parser.add_argument("--test", action=TestMode, nargs=0, help="run IsoQuant on toy dataset")
    add_hidden_option('--debug', action='store_true', default=False,
                      help='Debug log output.')

    output_args_group.add_argument("--output", "-o", help="output folder, will be created automatically "
                                                          "[default=isoquant_output]",
                                   type=str, default="isoquant_output")
    output_args_group.add_argument('--prefix', '-p', type=str,
                                   help='experiment name; to be used for folder and file naming; default is OUT',
                                   default="OUT")
    output_args_group.add_argument('--labels', '-l', nargs='+', type=str,
                                   help='sample/replica labels to be used as column names; input file names are used '
                                        'if not set; must be equal to the number of input files given via --fastq/--bam')
    # REFERENCE
    ref_args_group.add_argument("--reference", "-r", help="reference genome in FASTA format (can be gzipped)",
                                type=str)
    ref_args_group.add_argument("--genedb", "-g", help="gene database in gffutils DB format or GTF/GFF "
                                                       "format (optional)", type=str)
    ref_args_group.add_argument('--complete_genedb', action='store_true', default=False,
                                help="use this flag if gene annotation contains transcript and gene metafeatures, "
                                     "e.g. with official annotations, such as GENCODE; "
                                     "speeds up gene database conversion")
    add_additional_option_to_group(ref_args_group, "--discard_chr", nargs="+", help="chromosome IDs to ignore", type=str, default=[])
    add_additional_option_to_group(ref_args_group, "--process_only_chr", nargs="+", help="chromosome IDs to process",
                                   type=str, default=None)
    add_additional_option_to_group(ref_args_group, "--index", help="genome index for specified aligner (optional)",
                                   type=str)

    # INPUT READS

    input_args = input_args_group.add_mutually_exclusive_group()
    input_args.add_argument('--bam', nargs='+', type=str,
                            help='sorted and indexed BAM file(s), each file will be treated as a separate sample')
    input_args.add_argument('--fastq', nargs='+', type=str,
                            help='input FASTQ/FASTA file(s) with reads, each file will be treated as a separate sample; '
                                 'reference genome should be provided when using reads as input')
    input_args.add_argument('--unmapped_bam', nargs='+', type=str,
                            help='unmapped BAM file(s), each file will be treated as a separate sample')
    input_args.add_argument('--yaml', type=str, help='yaml file containing all input files, one entry per sample'
                                                     ', check readme for format info')

    input_args_group.add_argument('--illumina_bam', nargs='+', type=str,
                                  help='sorted and indexed file(s) with Illumina reads from the same sample')

    input_args_group.add_argument("--read_group", nargs='+', type=str,
                                  help="one or more ways to group feature counts (no grouping by default); "
                                       "multiple grouping strategies can be specified (space-separated); "
                                       "supported formats: "
                                       "tag:TAG (BAM tag), "
                                       "file:FILE:READ_COL:GROUP_COL(S):DELIM (TSV file, use comma-separated columns for multi-column grouping, e.g., file:table.tsv:0:1,2,3), "
                                       "read_id:DELIM (read ID suffix), "
                                       "file_name (original filename), "
                                       "barcode_spot (map barcodes to spots/cell types using --barcode2spot), "
                                       "barcode_barcode (map barcodes to spots using --barcode2barcode), "
                                       "barcode (group by barcode from --barcoded_reads)")

    add_additional_option_to_group(input_args_group, "--read_assignments", nargs='+', type=str,
                                   help="reuse read assignments (binary format)", default=None)

    # INPUT PROPERTIES
    input_args_group.add_argument("--data_type", "-d", type=str, choices=DATA_TYPE_ALIASES.keys(),
                        help="type of data to process, supported types are: " + ", ".join(DATA_TYPE_ALIASES.keys()))
    input_args_group.add_argument('--stranded',  type=str, help="reads strandness type, supported values are: " +
                        ", ".join(SUPPORTED_STRANDEDNESS), default="none")
    input_args_group.add_argument('--polya_trimmed', default=PolyATrimmed.none.name, type=str,
                                  choices=[e.name for e in PolyATrimmed],
                                  help="define reads which had polyA tail trimmed")
    input_args_group.add_argument('--fl_data', action='store_true', default=False,
                        help="reads represent FL transcripts; both ends of the read are considered to be reliable")

    # SC ARGUMENTS
    add_additional_option_to_group(sc_args_group, "--mode", "-m", type=str, choices=ISOQUANT_MODES,
                                   help="IsoQuant modes: " + ", ".join(ISOQUANT_MODES) +
                                        "; default:%s" % IsoQuantMode.bulk.name, default=IsoQuantMode.bulk.name)
    add_additional_option_to_group(sc_args_group, '--barcode_whitelist', type=str, nargs='+',
                                   help='file(s) with barcode whitelist(s) for barcode calling')
    add_additional_option_to_group(sc_args_group, "--barcoded_reads", type=str, nargs='+',
                                   help='TSV file(s) with barcoded reads; barcodes will be called automatically if not provided')
    add_additional_option_to_group(sc_args_group, "--barcoded_bam", action='store_true', default=False,
                                   help='extract barcodes and UMIs from BAM tags (CB/UB by default); '
                                        'bypasses barcode calling')
    add_additional_option_to_group(sc_args_group, "--barcode_tag", type=str, default="CB",
                                   help='BAM tag for cell barcode (default: CB)')
    add_additional_option_to_group(sc_args_group, "--umi_tag", type=str, default="UB",
                                   help='BAM tag for UMI (default: UB)')
    add_additional_option_to_group(sc_args_group, "--strip_barcode_suffix", action='store_true', default=False,
                                   help='remove suffix after dash from barcodes extracted from BAM tag (e.g. ACGT-1 -> ACGT)')
    add_additional_option_to_group(sc_args_group, "--barcode2spot", type=str,
                                   help='TSV file mapping barcode to cell type / spot id. '
                                        'Format: file.tsv or file.tsv:barcode_col:spot_cols '
                                        '(e.g., file.tsv:0:1,2,3 for multiple spot columns)')
    add_additional_option_to_group(sc_args_group, "--barcode2barcode", type=str,
                                   help='TSV file mapping barcode to spot IDs for UMI deduplication; '
                                        'format: file.tsv or file.tsv:barcode_col:spot_cols')
    add_additional_option_to_group(sc_args_group, "--molecule", type=str,
                                   help='molecule definition file (MDF) for custom_sc mode; '
                                        'defines molecule structure for universal barcode extraction')

    # ALGORITHM
    add_additional_option_to_group(algo_args_group, "--report_novel_unspliced", "-u", type=bool_str,
                                   help="report novel monoexonic transcripts (true/false), "
                                        "default: false for ONT, true for other data types")
    add_additional_option_to_group(algo_args_group, "--report_canonical",  type=str,
                                   choices=[e.name for e in StrandnessReportingLevel],
                                   help="reporting level for novel transcripts based on canonical splice sites;"
                                        " default: " + StrandnessReportingLevel.auto.name,
                                   default=StrandnessReportingLevel.auto.name)
    add_additional_option_to_group(algo_args_group, "--polya_requirement", type=str,
                                   choices=[e.name for e in PolyAUsageStrategies],
                                   help="require polyA tails to be present when reporting transcripts; "
                                        "default: auto (requires polyA only when polyA percentage is >= 70%%)",
                                   default=PolyAUsageStrategies.auto.name)

    add_additional_option_to_group(algo_args_group, "--transcript_quantification", choices=COUNTING_STRATEGIES,
                                   help="transcript quantification strategy", type=str,
                                   default=CountingStrategy.unique_only.name)
    add_additional_option_to_group(algo_args_group, "--gene_quantification", choices=COUNTING_STRATEGIES,
                                   help="gene quantification strategy", type=str,
                                   default=CountingStrategy.unique_splicing_consistent.name)

    add_additional_option_to_group(algo_args_group, "--matching_strategy",
                                   choices=["exact", "precise", "default", "loose"],
                                   help="read-to-isoform matching strategy from the most strict to least",
                                   type=str, default=None)
    add_additional_option_to_group(algo_args_group, "--splice_correction_strategy",
                                   choices=["none", "default_pacbio", "default_ont",
                                            "conservative_ont", "all", "assembly"],
                                   help="read alignment correction strategy to use", type=str, default=None)
    add_additional_option_to_group(algo_args_group, "--model_construction_strategy",
                                   choices=["reliable", "default_pacbio", "sensitive_pacbio", "fl_pacbio",
                                            "default_ont", "sensitive_ont", "all", "assembly"],
                                   help="transcript model construction strategy to use", type=str, default=None)
    add_additional_option_to_group(algo_args_group, "--delta", type=int, default=None,
                                   help="delta for inexact splice junction comparison, chosen automatically based on data type")
    add_additional_option_to_group(algo_args_group, "--use_replicas", type=bool_str, default=True,
                                   help="require novel transcripts to be confirmed by multiple files "
                                        "when file_name grouping is used (default: true)")


    # PIPELINE STEPS
    pipeline_args_group.add_argument("--threads", "-t", help="number of threads to use", type=int,
                                     default="16")

    resume_args = pipeline_args_group.add_mutually_exclusive_group()
    resume_args.add_argument("--resume", action="store_true", default=False,
                             help="resume failed run, specify output folder, input options are not allowed")
    resume_args.add_argument("--force", action="store_true", default=False,
                             help="force to overwrite the previous run")

    add_additional_option_to_group(pipeline_args_group, '--clean_start', action='store_true', default=False,
                                   help='Do not use previously generated index, feature db or alignments.')
    add_additional_option_to_group(pipeline_args_group, "--no_model_construction", action="store_true",
                                   default=False, help="run only read assignment and quantification")
    add_additional_option_to_group(pipeline_args_group, "--run_aligner_only", action="store_true", default=False,
                                   help="align reads to reference without running further analysis")
    add_additional_option_to_group(pipeline_args_group, "--no_gtf_check", help="do not perform GTF checks",
                                   dest="gtf_check",
                                   action='store_false', default=True)
    add_additional_option_to_group(pipeline_args_group, "--high_memory",
                                   help="increase RAM consumption (store alignment and the genome in RAM)",
                                   action='store_true', default=False)
    add_additional_option_to_group(pipeline_args_group, "--keep_tmp", help="do not remove temporary files "
                                                                           "in the end", action='store_true',
                                   default=False)

    # OUTPUT SETUP
    output_setup_args_group.add_argument('--check_canonical', action='store_true', default=False,
                                     help="report whether splice junctions are canonical")
    output_setup_args_group.add_argument("--sqanti_output", help="produce SQANTI-like TSV output",
                                     action='store_true', default=False)
    output_setup_args_group.add_argument("--count_exons", help="perform exon and intron counting",
                                     action='store_true', default=False)
    add_additional_option_to_group(output_setup_args_group,"--bam_tags",
                                   help="comma separated list of BAM tags to be imported to read_assignments.tsv",
                                   type=str)
    add_additional_option_to_group(output_setup_args_group, "--no_gzip", help="do not gzip large output files", dest="gzipped",
                                   action='store_false', default=True)
    add_additional_option_to_group(output_setup_args_group, "--large_output", nargs='*', type=str,
                                   default=["read_assignments", "allinfo"],
                                   help="large output files to generate: " + ", ".join(LARGE_OUTPUT_TYPES) +
                                        " (default: read_assignments allinfo)")
    add_additional_option_to_group(output_setup_args_group, "--normalization_method", type=str, choices=[e.name for e in NormalizationMethod],
                                   help="TPM normalization method: simple - conventional normalization using all counted reads;"
                                        "usable_reads - includes all assigned reads;"
                                        "none - do not convert counts to TPM.",
                                   default=NormalizationMethod.simple.name)
    add_additional_option_to_group(output_setup_args_group, "--counts_format", type=str, nargs='+',
                                   choices=[e.name for e in GroupedOutputFormat],
                                   help="output format for grouped counts",
                                   default=[GroupedOutputFormat.default.name])

    add_additional_option_to_group(output_setup_args_group, "--genedb_output", help="output folder for converted gene "
                                                                                    "database, will be created automatically "
                                                                                    " (same as output by default)", type=str)

    # ALIGNER
    add_additional_option_to_group(align_args_group, "--aligner", help="force to use this alignment method, can be " + ", ".join(SUPPORTED_ALIGNERS)
                                        + "; chosen based on data type if not set", type=str)
    add_additional_option_to_group(align_args_group,  "--no_junc_bed", action="store_true", default=False,
                          help="do NOT use annotation for read mapping")
    add_additional_option_to_group(align_args_group, "--junc_bed_file", type=str,
                          help="annotation in BED format produced by minimap's paftools.js gff2bed "
                               "(will be created automatically if not given)")
    add_additional_option_to_group(align_args_group, "--indexing_options", type=str,
                                   help="additional options that will be passed to the aligner indexer")
    add_additional_option_to_group(align_args_group, "--mapping_options", type=str,
                                   help="additional options that will be passed to the aligner")

    # READ FILTERING
    add_additional_option_to_group(filer_args_group, "--use_secondary",
                                   help="use secondary alignments (slower processing)",
                                   action='store_true', default=False)
    add_additional_option_to_group(filer_args_group, "--no_secondary",
                                   help="deprecated, secondary alignments are not used by default (kept for user convenience)",
                                   action='store_true', default=False)
    add_additional_option_to_group(filer_args_group, "--min_mapq",
                                   help="ignore alignments with MAPQ < this (also filters out secondary alignments, default: None)", type=int)
    add_additional_option_to_group(filer_args_group, "--inconsistent_mapq_cutoff",
                                   help="ignore inconsistent alignments with MAPQ < this (works only with the reference annotation, default=5)",
                                   type=int, default=5)
    add_additional_option_to_group(filer_args_group, "--simple_alignments_mapq_cutoff",
                                   help="ignore alignments with 1 or 2 exons and MAPQ < this "
                                        "(works only in annotation-free mode, default=1)", type=int, default=1)
    add_additional_option_to_group(filer_args_group, "--max_coverage_small_chr",
                                   help="process only a fraction of reads for high-coverage loci on small chromosomes, "
                                        "e.g. mitochondrial (default 1000000); significantly improves running time and RAM",
                                   type=int, default=1000000)
    add_additional_option_to_group(filer_args_group, "--max_coverage_normal_chr",
                                   help="process only a fraction of reads for high-coverage loci on usual chromosomes"
                                        " (default -1 = infinity);  improves running time and RAM",
                                   type=int, default=-1)

    # REST
    add_hidden_option("--graph_clustering_distance", type=int, default=None,
                      help="intron graph clustering distance, "
                           "splice junctions less that this number of bp apart will not be differentiated")
    add_hidden_option("--cage", help="bed file with CAGE peaks", type=str, default=None)
    add_hidden_option("--cage-shift", type=int, default=50, help="interval before read start to look for CAGE peak")

    isoquant_version = "3.12.0"
    try:
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "VERSION")) as version_f:
            isoquant_version = version_f.readline().strip()
    except FileNotFoundError:
        try:
            from importlib.metadata import version as _get_version
            isoquant_version = _get_version("isoquant")
        except Exception:
            pass
    parser.add_argument('--version', '-v', action='version', version='IsoQuant ' + isoquant_version)

    args = parser.parse_args(cmd_args, namespace)

    if args.resume:
        resume_parser = argparse.ArgumentParser(add_help=False)
        resume_parser.add_argument("--resume", action="store_true", default=False,
                                   help="resume failed run, specify only output folder, "
                                        "input options are not allowed")
        resume_parser.add_argument("--output", "-o",
                                   help="output folder, will be created automatically [default=isoquant_output]",
                                   type=str, required=True)
        resume_parser.add_argument('--debug', action='store_true', default=argparse.SUPPRESS,
                                   help='Debug log output.')
        resume_parser.add_argument("--threads", "-t", help="number of threads to use",
                                   type=int, default=argparse.SUPPRESS)
        resume_parser.add_argument("--high_memory",
                                   help="increase RAM consumption (store alignment and the genome in RAM)",
                                   action='store_true', default=False)
        resume_parser.add_argument("--keep_tmp", help="do not remove temporary files in the end",
                                   action='store_true', default=argparse.SUPPRESS)

        args, unknown_args = resume_parser.parse_known_args(cmd_args)
        if unknown_args:
            logger.error("You cannot specify options other than --output/--threads/--debug/--high_memory "
                         "with --resume option")
            parser.print_usage()
            sys.exit(IsoQuantExitCode.INCOMPATIBLE_OPTIONS)

    args._cmd_line = " ".join(sys.argv)
    args._version = isoquant_version

    args.output_exists = os.path.exists(args.output)
    if not args.output_exists:
        os.makedirs(args.output)

    return args, parser


def check_and_load_args(args, parser):
    args.param_file = os.path.join(args.output, ".params")
    if args.resume:
        if not os.path.exists(args.output) or not os.path.exists(args.param_file):
            # logger is not defined yet
            logger.error("Previous run config was not detected, cannot resume. "
                         "Check that output folder is correctly specified.")
            sys.exit(IsoQuantExitCode.RESUME_CONFIG_NOT_FOUND)
        args = load_previous_run(args)
    elif args.output_exists:
        if os.path.exists(args.param_file):
            if args.force:
                logger.warning("Output folder already contains a previous run, will be overwritten.")
            else:
                logger.warning("Output folder already contains a previous run, some files may be overwritten. "
                               "Use --resume to resume a failed run. Use --force to avoid this message.")
                logger.warning("Press Ctrl+C to interrupt the run now.")
                delay = 9
                for i in range(delay):
                    countdown = delay - i
                    sys.stdout.write("Resuming the run in %d second%s\r" % (countdown, "s" if countdown > 1 else ""))
                    time.sleep(1)
                logger.info("Overwriting the previous run")
                time.sleep(1)
        else:
            logger.warning("Output folder already exists, some files may be overwritten.")

    if args.genedb_output is None:
        args.genedb_output = args.output
    elif not os.path.exists(args.genedb_output):
        os.makedirs(args.genedb_output)
    if not args.genedb:
        args.genedb_filename = None
    elif args.genedb.lower().endswith("db"):
        args.genedb_filename = args.genedb
    else:
        args.genedb_filename = os.path.join(args.genedb_output, os.path.splitext(os.path.basename(args.genedb))[0] + ".db")

    if not check_input_params(args):
        parser.print_usage()
        sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

    # Validate --read_group none
    if args.read_group:
        if "none" in args.read_group and len(args.read_group) > 1:
            logger.error("--read_group 'none' cannot be combined with other values")
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

    # Validate --large_output values
    if args.large_output:
        if "none" in args.large_output and len(args.large_output) > 1:
            logger.error("--large_output 'none' cannot be combined with other values")
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
        for val in args.large_output:
            if val not in LARGE_OUTPUT_TYPES:
                logger.error("Invalid --large_output value: %s. Valid values: %s" % (val, ", ".join(LARGE_OUTPUT_TYPES)))
                sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

    save_params(args)
    return args


def load_previous_run(args):
    logger.info("Loading parameters from the previous run")
    logger.error("Only --output/--threads/--debug/--high_memory are compatible with --resume option")
    unpickler = pickle.Unpickler(open(args.param_file, "rb"), fix_imports=False)
    loaded_args = unpickler.load()

    for option in args.__dict__:
        loaded_args.__dict__[option] = args.__dict__[option]

    if loaded_args.debug:
        logger.setLevel(logging.DEBUG)
        logger.handlers[0].setLevel(logging.DEBUG)

    return loaded_args


def save_params(args):
    for file_opt in ["genedb", "reference", "index", "bam", "fastq", "junc_bed_file",
                     "cage", "genedb_output", "read_assignments"]:
        if file_opt in args.__dict__ and args.__dict__[file_opt]:
            if isinstance(args.__dict__[file_opt], list):
                args.__dict__[file_opt] = list(map(os.path.abspath, args.__dict__[file_opt]))
            else:
                args.__dict__[file_opt] = os.path.abspath(args.__dict__[file_opt])

    if "read_group" in args.__dict__ and args.__dict__["read_group"]:
        updated_specs = []
        for spec in args.read_group:
            vals = spec.split(":")
            if len(vals) > 1 and vals[0] == 'file':
                vals[1] = os.path.abspath(vals[1])
                updated_specs.append(":".join(vals))
            else:
                updated_specs.append(spec)
        args.read_group = updated_specs

    pickler = pickle.Pickler(open(args.param_file, "wb"),  -1)
    pickler.dump(args)
    pass


# Check user's params
def check_input_params(args):
    if not args.reference:
        logger.error("Reference genome was not provided")
        return False
    if not args.data_type:
        logger.error("Data type is not provided, choose one of " + " ".join(DATA_TYPE_ALIASES.keys()))
        return False
    elif args.data_type not in DATA_TYPE_ALIASES.keys():
        logger.error("Unsupported data type " + args.data_type + ", choose one of: " + " ".join(DATA_TYPE_ALIASES.keys()))
        return False
    args.data_type = DATA_TYPE_ALIASES[args.data_type]

    if not args.fastq and not args.bam and not args.unmapped_bam and not args.read_assignments and not args.yaml:
        logger.error("No input data was provided")
        return False

    if args.yaml and args.illumina_bam:
        logger.error("When providing a yaml file it should include all input files, including the illumina bam file.")
        return False

    args.input_data = InputDataStorage(args)
    if args.aligner is not None and args.aligner not in SUPPORTED_ALIGNERS:
        logger.error(" Unsupported aligner " + args.aligner + ", choose one of: " + " ".join(SUPPORTED_ALIGNERS))
        return False

    if args.run_aligner_only and not args.input_data.input_type.needs_mapping():
        logger.error("Data type %s cannot be mapped and thus incompatible with --run_aligner_only option." % args.input_data.input_type.name)
        return False
    if args.stranded not in SUPPORTED_STRANDEDNESS:
        logger.error("Unsupported strandness " + args.stranded + ", choose one of: " + " ".join(SUPPORTED_STRANDEDNESS))
        return False

    if not args.genedb:
        if args.count_exons:
            logger.warning("--count_exons option has no effect without gene annotation")
        if args.sqanti_output:
            args.sqanti_output = False
            logger.warning("--sqanti_output option has no effect without gene annotation")
        if args.no_model_construction:
            logger.warning("Setting --no_model_construction without providing a gene "
                           "annotation will not produce any meaningful results")

    if args.no_model_construction and args.sqanti_output:
        args.sqanti_output = False
        logger.warning("--sqanti_output option has no effect without model construction")

    if args.no_secondary:
        logger.info("--no_secondary option has no effect and will be deprecated, secondary alignments are not used by default")

    if args.process_only_chr and args.discard_chr:
        args.discard_chr = []
        logger.warning("--discard_chr has not effect when --process_only_chr is set and will be ignored")

    if "read_group" in args.__dict__ and args.__dict__["read_group"]:
        updated_specs = []
        spec_set = set()
        for spec in args.read_group:
            if spec in spec_set:
                logger.warning("Read group %s is set twice, which has no effect, duplicate will be ignored" % spec)
                continue
            updated_specs.append(spec)
            spec_set.add(spec)
        args.read_group = updated_specs

    if not isinstance(args.mode, IsoQuantMode):
        args.mode = IsoQuantMode[args.mode]

    args.umi_length = 0
    if args.mode.needs_barcode_calling():
        barcode_sources = sum([bool(args.barcode_whitelist), bool(args.barcoded_reads), bool(args.barcoded_bam)])
        if barcode_sources > 1:
            logger.critical("Options --barcode_whitelist, --barcoded_reads, and --barcoded_bam are mutually exclusive")
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
        if args.mode == IsoQuantMode.custom_sc:
            if not args.molecule and not args.barcoded_reads and not args.barcoded_bam:
                logger.critical("custom_sc mode requires --molecule, --barcoded_reads, or --barcoded_bam")
                sys.exit(IsoQuantExitCode.BARCODE_WHITELIST_MISSING)
        elif not args.barcode_whitelist and not args.barcoded_reads and not args.barcoded_bam:
            logger.critical("You have chosen single-cell/spatial mode %s, please specify barcode whitelist, "
                            "file with barcoded reads, or --barcoded_bam" % args.mode.name)
            sys.exit(IsoQuantExitCode.BARCODE_WHITELIST_MISSING)
        if args.barcoded_bam:
            args.umi_length = _detect_umi_length_from_bam(args.input_data.samples[0].file_list[0][0], args.umi_tag)
        else:
            args.umi_length = get_umi_length(args.mode)

    check_input_files(args)
    return True


def _detect_umi_length_from_bam(bam_path: str, umi_tag: str) -> int:
    with pysam.AlignmentFile(bam_path, "rb") as bam:
        for read in bam:
            if read.has_tag(umi_tag):
                return len(read.get_tag(umi_tag))
    return 0


def check_bam_file(bam_path: str, check_index: bool = True):
    """Check BAM file exists and optionally has index."""
    if not os.path.isfile(bam_path):
        logger.critical("BAM file " + bam_path + " does not exist")
        sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)
    if check_index:
        bamfile_in = pysam.AlignmentFile(bam_path, "rb")
        if not bamfile_in.has_index():
            logger.critical("BAM file " + bam_path + " is not indexed, run samtools sort and samtools index")
            sys.exit(IsoQuantExitCode.BAM_NOT_INDEXED)
        bamfile_in.close()


def check_file_exists(file_path: str, description: str):
    """Check that a file exists, exit with error if not."""
    if not os.path.isfile(file_path):
        logger.critical(f"{description} {file_path} does not exist")
        sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)


def extract_read_group_file_path(spec: str):
    """
    Extract file path from read_group spec if it's a file-based spec.

    Returns file path for 'file:path:...' specs, None otherwise.
    """
    parts = spec.split(":")
    if len(parts) >= 2 and parts[0] == 'file':
        return parts[1]
    return None


def check_input_files(args):
    # Check reference genome
    if args.reference and not os.path.isfile(args.reference):
        logger.critical("Reference genome " + args.reference + " does not exist")
        sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)

    # Check input reads (BAM/FASTQ/save files)
    for sample in args.input_data.samples:
        for lib in sample.file_list:
            for in_file in lib:
                if args.input_data.input_type == InputDataType.save:
                    saves = glob.glob(in_file + "*")
                    if not saves:
                        logger.critical("Input files " + in_file + "* do not exist")
                    continue
                if not os.path.isfile(in_file):
                    logger.critical("Input file " + in_file + " does not exist")
                    sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)
                if args.input_data.input_type == InputDataType.bam:
                    check_bam_file(in_file, check_index=True)

        # Check Illumina BAM files
        if sample.illumina_bam is not None:
            for illumina in sample.illumina_bam:
                check_bam_file(illumina, check_index=True)

    # Check barcoded reads files (from args, not sample - sample.barcoded_reads is set later)
    if hasattr(args, 'barcoded_reads') and args.barcoded_reads:
        if isinstance(args.barcoded_reads, list):
            for bc_file in args.barcoded_reads:
                check_file_exists(bc_file, "Barcoded reads file")
        else:
            check_file_exists(args.barcoded_reads, "Barcoded reads file")

    # Check molecule definition file
    if hasattr(args, 'molecule') and args.molecule:
        check_file_exists(args.molecule, "Molecule definition file")

    # Check barcode whitelist files
    if hasattr(args, 'barcode_whitelist') and args.barcode_whitelist:
        for wl_file in args.barcode_whitelist:
            check_file_exists(wl_file, "Barcode whitelist file")

    # Check barcode2spot file (parse spec to extract filename)
    if hasattr(args, 'barcode2spot') and args.barcode2spot:
        from src.read_groups import parse_barcode2spot_spec
        bc2spot_file, _, _ = parse_barcode2spot_spec(args.barcode2spot)
        check_file_exists(bc2spot_file, "Barcode to spot mapping file")

    # Check barcode2barcode file (parse spec to extract filename)
    if hasattr(args, 'barcode2barcode') and args.barcode2barcode:
        from src.read_groups import parse_barcode2spot_spec
        bc2bc_file, _, _ = parse_barcode2spot_spec(args.barcode2barcode)
        check_file_exists(bc2bc_file, "Barcode to barcode mapping file")

    # Check read_group file specs
    if hasattr(args, 'read_group') and args.read_group:
        for spec in args.read_group:
            file_path = extract_read_group_file_path(spec)
            if file_path:
                check_file_exists(file_path, "Read group file")

    # Check junction BED file
    if hasattr(args, 'junc_bed_file') and args.junc_bed_file:
        check_file_exists(args.junc_bed_file, "Junction BED file")

    # Check CAGE file (currently not supported)
    if args.cage is not None:
        logger.critical("CAGE data is not supported yet")
        sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
        if not os.path.isfile(args.cage):
            logger.critical("Bed file with CAGE peaks " + args.cage + " does not exist")
            sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)

    # Check gene database
    if args.genedb is not None:
        if not os.path.isfile(args.genedb):
            logger.critical("Gene database " + args.genedb + " does not exist")
            sys.exit(IsoQuantExitCode.GENE_DB_NOT_FOUND)
    else:
        args.no_junc_bed = True

    # Check read assignments
    if args.read_assignments is not None:
        for r in args.read_assignments:
            if not glob.glob(r + "*"):
                logger.critical("No files found with prefix " + str(r))
                sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)


def create_output_dirs(args):
    for sample in args.input_data.samples:
        sample_dir = sample.out_dir
        if os.path.exists(sample_dir):
            if not args.resume:
                logger.warning(sample_dir + " folder already exists, some files may be overwritten")
        else:
            os.makedirs(sample_dir)
        sample_aux_dir = sample.aux_dir
        if os.path.exists(sample_aux_dir):
            if not args.resume:
                logger.warning(sample_aux_dir + " folder already exists, some files may be overwritten")
        else:
            os.makedirs(sample_aux_dir)


def set_logger(args):
    output_level = logging.DEBUG if args.__dict__.get('debug') else logging.INFO
    log_file = os.path.join(args.output, "isoquant.log")
    if os.path.exists(log_file):
        old_log_file = os.path.join(args.output, "isoquant.log.old")
        with open(old_log_file, "a") as olf:
            olf.write("\n")
            shutil.copyfileobj(open(log_file, "r"), olf)
    with open(log_file, "w") as f:
        f.write("Command line: " + args._cmd_line + '\n')
    setup_worker_logging(log_file, output_level)
    logger.info("Running IsoQuant version " + args._version)


def set_data_dependent_options(args):
    matching_strategies = {ASSEMBLY: "precise", PACBIO_CCS_DATA: "precise", NANOPORE_DATA: "default"}
    if args.matching_strategy is None:
        args.matching_strategy = matching_strategies[args.data_type]

    model_construction_strategies = {ASSEMBLY: "assembly", PACBIO_CCS_DATA: "default_pacbio", NANOPORE_DATA: "default_ont"}
    if args.model_construction_strategy is None:
        args.model_construction_strategy = model_construction_strategies[args.data_type]
        if args.fl_data and args.model_construction_strategy == "default_pacbio":
            args.model_construction_strategy = "fl_pacbio"

    splice_correction_strategies = {ASSEMBLY: "assembly", PACBIO_CCS_DATA: "default_pacbio", NANOPORE_DATA: "default_ont"}
    if args.splice_correction_strategy is None:
        args.splice_correction_strategy = splice_correction_strategies[args.data_type]

    args.resolve_ambiguous = 'monoexon_and_fsm' if args.fl_data else 'default'
    args.requires_polya_for_construction = False

    # Handle --read_group none: disable all auto-added groupings
    if args.read_group is not None and "none" in args.read_group:
        args.read_group = None
        return

    # Automatically add file_name grouping when multiple files are present
    if args.input_data.has_replicas():
        if args.read_group is None:
            # No read grouping specified, use file_name
            args.read_group = ["file_name"]
        else:
            # Read grouping specified, ensure file_name is included
            if "file_name" not in args.read_group:
                args.read_group.append("file_name")

    # Automatically add barcode_spot grouping when --barcode2spot is set
    if hasattr(args, 'barcode2spot') and args.barcode2spot:
        if args.read_group is None:
            args.read_group = ["barcode_spot"]
        elif "barcode_spot" not in args.read_group:
            args.read_group.append("barcode_spot")

    # Automatically add barcode_barcode grouping when --barcode2barcode is set
    if hasattr(args, 'barcode2barcode') and args.barcode2barcode:
        if args.read_group is None:
            args.read_group = ["barcode_barcode"]
        elif "barcode_barcode" not in args.read_group:
            args.read_group.append("barcode_barcode")

    # In SC modes, auto-add barcode grouping if no barcode-related grouping is set
    if args.mode.needs_barcode_calling():
        barcode_groupings = {"barcode", "barcode_spot", "barcode_barcode"}
        has_barcode_grouping = (args.read_group is not None and
                                any(rg in barcode_groupings for rg in args.read_group))
        if not has_barcode_grouping:
            if args.read_group is None:
                args.read_group = ["barcode"]
            else:
                args.read_group.append("barcode")
            logger.info("Single-cell/spatial mode: automatically adding '--read_group barcode'. "
                        "Use '--read_group none' to disable.")


def set_matching_options(args):
    MatchingStrategy = namedtuple('MatchingStrategy',
                                  ('delta', 'max_intron_shift', 'max_missed_exon_len', 'max_fake_terminal_exon_len',
                                   'max_suspicious_intron_abs_len', 'max_suspicious_intron_rel_len',
                                   'resolve_ambiguous', 'correct_minor_errors'))

    strategies = {
        'exact':   MatchingStrategy(0, 0, 0, 0, 0, 0.0, 'monoexon_only', False),
        'precise': MatchingStrategy(4, 30, 50, 20, 0, 0.0, 'monoexon_and_fsm', True),
        'default': MatchingStrategy(6, 60, 100, 40, 60, 1.0, 'monoexon_and_fsm', True),
        'loose':   MatchingStrategy(12, 60, 100, 40, 60, 1.0, 'all',  True),
    }

    strategy = strategies[args.matching_strategy]

    if args.delta is None:
        args.delta = strategy.delta
    elif args.delta < 0:
        logger.error("--delta can not be negative")
        sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
    args.minor_exon_extension = 50
    args.major_exon_extension = 300
    args.max_intron_shift = strategy.max_intron_shift
    args.max_missed_exon_len = strategy.max_missed_exon_len
    args.max_fake_terminal_exon_len = strategy.max_fake_terminal_exon_len
    # short introns that are actually long deletions, fix minimaps logic
    args.max_suspicious_intron_abs_len = strategy.max_suspicious_intron_abs_len
    args.max_suspicious_intron_rel_len = strategy.max_suspicious_intron_rel_len
    args.min_abs_exon_overlap = 10
    args.min_rel_exon_overlap = 0.2
    args.micro_intron_length = 50
    args.max_intron_abs_diff = min(30, args.max_intron_shift)
    args.max_intron_rel_diff = 0.2
    args.apa_delta = args.minor_exon_extension
    args.minimal_exon_overlap = 5
    args.minimal_intron_absence_overlap = 20
    args.polya_window = 16
    args.polya_fraction = 0.75
    if args.resolve_ambiguous == 'default':
        args.resolve_ambiguous = strategy.resolve_ambiguous
    if args.resolve_ambiguous not in AmbiguityResolvingMethod.__dict__:
        logger.error("Incorrect resolving ambiguity method: " + args.resolve_ambiguous + ", default will be used")
        args.resolve_ambiguous = strategy.resolve_ambiguous
    args.resolve_ambiguous = AmbiguityResolvingMethod[args.resolve_ambiguous]
    args.correct_minor_errors = strategy.correct_minor_errors

    updated_strategy = MatchingStrategy(args.delta, args.max_intron_shift, args.max_missed_exon_len,
                                        args.max_fake_terminal_exon_len,
                                        args.max_suspicious_intron_abs_len, args.max_suspicious_intron_rel_len,
                                        args.resolve_ambiguous, args.correct_minor_errors)
    logger.debug('Using %s strategy. Updated strategy: %s.' % (args.matching_strategy, updated_strategy))


def set_splice_correction_options(args):
    SplicSiteCorrectionStrategy = namedtuple('SplicSiteCorrectionStrategy',
                                             ('fuzzy_junctions', 'intron_shifts', 'skipped_exons',
                                              'terminal_exons', 'fake_terminal_exons', 'microintron_retention'))
    strategies = {
        'none': SplicSiteCorrectionStrategy(False, False, False, False, False, False),
        'default_pacbio': SplicSiteCorrectionStrategy(True, False, True, False, False, True),
        'conservative_ont': SplicSiteCorrectionStrategy(True, False, True, False, False, False),
        'default_ont': SplicSiteCorrectionStrategy(True, False, True, False, True, True),
        'all': SplicSiteCorrectionStrategy(True, True, True, True, True, True),
        'assembly': SplicSiteCorrectionStrategy(False, False, True, False, False, False)
    }
    strategy = strategies[args.splice_correction_strategy]
    args.correct_fuzzy_junctions = strategy.fuzzy_junctions
    args.correct_intron_shifts = strategy.intron_shifts
    args.correct_skipped_exons = strategy.skipped_exons
    args.correct_terminal_exons = strategy.terminal_exons
    args.correct_fake_terminal_exons = strategy.fake_terminal_exons
    args.correct_microintron_retention = strategy.microintron_retention


def set_model_construction_options(args):
    ModelConstructionStrategy = namedtuple('ModelConstructionStrategy',
                                           ('min_novel_intron_count',
                                            'graph_clustering_ratio', 'graph_clustering_distance',
                                            'min_novel_isolated_intron_abs', 'min_novel_isolated_intron_rel',
                                            'terminal_position_abs', 'terminal_position_rel',
                                            'terminal_internal_position_rel',
                                            'min_known_count', 'min_nonfl_count',
                                            'min_novel_count', 'min_novel_count_rel',
                                            'min_mono_count_rel', 'singleton_adjacent_cov',
                                            'fl_only', 'novel_monoexonic',
                                            'require_monointronic_polya', 'require_monoexonic_polya',
                                            'report_canonical'))
    strategies = {
        'reliable':        ModelConstructionStrategy(2, 0.5, 20,  5, 0.05,  1, 0.1,  0.1,  2, 4, 8, 0.05, 0.05, 50,
                                                     True, False, True, True, StrandnessReportingLevel.only_canonical),
        'default_pacbio':  ModelConstructionStrategy(1, 0.5, 10,  2, 0.02,  1, 0.05,  0.05,  1, 2, 2, 0.02, 0.005, 100,
                                                     False, True, False, True, StrandnessReportingLevel.only_canonical),
        'sensitive_pacbio':ModelConstructionStrategy(1, 0.5, 5,   2, 0.005,  1, 0.01,  0.02,  1, 2, 2, 0.005, 0.001, 100,
                                                     False, True, False, False, StrandnessReportingLevel.only_stranded),
        'default_ont':     ModelConstructionStrategy(1, 0.5, 20,  3, 0.02,  1, 0.05,  0.05,  1, 3, 3, 0.02, 0.02, 10,
                                                     False, False, True, True, StrandnessReportingLevel.only_canonical),
        'sensitive_ont':   ModelConstructionStrategy(1, 0.5, 20,  3, 0.005,  1, 0.01,  0.02,  1, 2, 3, 0.005, 0.005, 10,
                                                     False, True, False, False, StrandnessReportingLevel.only_stranded),
        'fl_pacbio':       ModelConstructionStrategy(1, 0.5, 10,  2, 0.02,  1, 0.05,  0.01,  1, 2, 3, 0.02, 0.005, 100,
                                                     True, True, False, False, StrandnessReportingLevel.only_canonical),
        'all':             ModelConstructionStrategy(0, 0.3, 5,   1, 0.002,  1, 0.01, 0.01, 1, 1, 1, 0.002, 0.001, 500,
                                                     False, True, False, False, StrandnessReportingLevel.all),
        'assembly':        ModelConstructionStrategy(0, 0.3, 5,   1, 0.05,  1, 0.01, 0.02,  1, 1, 1, 0.05, 0.01, 50,
                                                     False, True, False, False, StrandnessReportingLevel.only_stranded)
    }
    strategy = strategies[args.model_construction_strategy]

    args.min_novel_intron_count = strategy.min_novel_intron_count
    args.graph_clustering_ratio = strategy.graph_clustering_ratio
    if args.graph_clustering_distance is None:
        args.graph_clustering_distance = strategy.graph_clustering_distance
    elif args.graph_clustering_distance < 0:
        logger.error("--graph_clustering_distance can not be negative")
        sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
    args.min_novel_isolated_intron_abs = strategy.min_novel_isolated_intron_abs
    args.min_novel_isolated_intron_rel = strategy.min_novel_isolated_intron_rel
    args.terminal_position_abs = strategy.terminal_position_abs
    args.terminal_position_rel = strategy.terminal_position_rel
    args.terminal_internal_position_rel = strategy.terminal_internal_position_rel

    args.min_known_count = strategy.min_known_count
    args.min_nonfl_count = strategy.min_nonfl_count
    args.min_novel_count = strategy.min_novel_count
    args.min_mono_count_rel = strategy.min_mono_count_rel
    args.min_novel_count_rel = strategy.min_novel_count_rel
    args.singleton_adjacent_cov = strategy.singleton_adjacent_cov
    args.fl_only = strategy.fl_only
    args.min_mono_exon_coverage = 0.75

    if args.report_novel_unspliced is None:
        args.report_novel_unspliced = strategy.novel_monoexonic

    if not args.report_novel_unspliced and not args.no_model_construction:
        logger.info("Novel unspliced transcripts will not be reported, "
                    "set --report_novel_unspliced true to discover them")

    args.require_monointronic_polya = strategy.require_monointronic_polya
    args.require_monoexonic_polya = strategy.require_monoexonic_polya
    args.polya_requirement_strategy = PolyAUsageStrategies[args.polya_requirement]
    args.polya_trimmed = PolyATrimmed[args.polya_trimmed]
    args.report_canonical_strategy = StrandnessReportingLevel[args.report_canonical]
    if args.report_canonical_strategy == StrandnessReportingLevel.auto:
        args.report_canonical_strategy = strategy.report_canonical


def set_configs_directory(args):
    config_dir = os.path.join(os.environ['HOME'], '.config', 'IsoQuant')
    os.makedirs(config_dir, exist_ok=True)

    args.db_config_path = os.path.join(config_dir, 'db_config.json')
    args.index_config_path = os.path.join(config_dir, 'index_config.json')
    args.bed_config_path = os.path.join(config_dir, 'bed_config.json')
    args.alignment_config_path = os.path.join(config_dir, 'alignment_config.json')
    for config_path in (args.db_config_path, args.index_config_path, args.bed_config_path, args.alignment_config_path):
        if not os.path.exists(config_path):
            with open(config_path, 'w') as f_out:
                json.dump({}, f_out, indent=2)


def set_additional_params(args):
    set_configs_directory(args)
    set_data_dependent_options(args)
    set_matching_options(args)
    set_model_construction_options(args)
    set_splice_correction_options(args)

    args.print_additional_info = True
    args.indel_near_splice_site_dist = 10
    args.upstream_region_len = 20

    args.multimap_strategy = "take_best"
    multimap_strategies = {}
    for e in MultimapResolvingStrategy:
        multimap_strategies[e.name] = e.value
    args.multimap_strategy = MultimapResolvingStrategy(multimap_strategies[args.multimap_strategy])

    args.needs_reference = True
    if args.needs_reference and not args.reference:
        logger.warning("Reference genome is not provided! This may affect quality of the results!")
        args.needs_reference = False

    args.simple_models_mapq_cutoff = 30
    args.polya_percentage_threshold = 0.7
    args.low_polya_percentage_threshold = 0.1

    if args.bam_tags:
        args.bam_tags = args.bam_tags.split(",")
    else:
        args.bam_tags = []
    args.original_annotation = None


def prepare_reference_genome(args):
    if not args.needs_reference:
        return
    logger.info("Reading reference genome from %s" % args.reference)
    ref_dir = os.path.dirname(args.reference)
    ref_file_name = os.path.basename(args.reference)
    ref_name, outer_ext = os.path.splitext(ref_file_name)

    # make symlink for pyfaidx index
    args.fai_file_name = args.reference + ".fai"
    if not os.path.exists(args.fai_file_name) and not os.access(ref_dir, os.W_OK):
        # index does not exist near the reference and reference folder is not writable
        # store index in the output folder in this case
        args.fai_file_name = os.path.join(args.output, ref_file_name + ".fai")

    low_ext = outer_ext.lower()
    if low_ext in ['.gz', '.gzip', '.bgz']:
        gunzipped_reference = os.path.join(args.output, ref_name)
        if not os.path.exists(gunzipped_reference) or not args.resume:
            logger.info("Decompressing reference to " + str(gunzipped_reference))
            with open(gunzipped_reference, "w") as outf:
                shutil.copyfileobj(gzip.open(args.reference, "rt"), outf)
        args.reference = gunzipped_reference


class BarcodeCallingArgs:
    def __init__(self, input, barcode_whitelist, mode, output, out_fasta, tmp_dir, threads,
                 molecule: str = None):
        self.input = input  # Can be a single file (str) or list of files
        self.barcodes = barcode_whitelist
        self.mode = mode
        self.output_tsv = output  # Can be a single filename (str) or list of filenames
        self.out_fasta = out_fasta  # Can be a single filename (str), list of filenames, or None
        self.tmp_dir = tmp_dir
        self.threads = threads
        self.molecule = molecule


def call_barcodes(args):
    if args.barcoded_bam:
        logger.info("Barcodes will be extracted from BAM tags (%s/%s)" % (args.barcode_tag, args.umi_tag))
        return
    if args.barcoded_reads:
        # TODO barcoded files via YAML
        args.input_data.samples[0].barcoded_reads = args.barcoded_reads
        return
    for sample in args.input_data.samples:
        # Collect all input files for this sample
        input_files = [files[0] for files in sample.file_list]
        output_barcodes_list = [sample.barcodes_tsv + "_%d.tsv" % i for i in range(len(input_files))]
        barcodes_done_list = [sample.barcodes_done + "_%d.tsv" % i for i in range(len(input_files))]

        output_fasta_list = None
        new_reads = []
        if args.mode.produces_new_fasta():
            output_fasta_list = [sample.split_reads_fasta + "_%d.fa" % i for i in range(len(input_files))]
            new_reads = [[fasta] for fasta in output_fasta_list]

        # Check if all files were already processed during resume
        all_done = all(os.path.exists(done) for done in barcodes_done_list)
        if all_done and args.resume:
            logger.info("Barcodes were called during the previous run, skipping")
            sample.barcoded_reads.extend(output_barcodes_list)
            if args.mode.produces_new_fasta():
                sample.file_list = new_reads
            continue

        # Remove existing done markers
        for barcodes_done in barcodes_done_list:
            if os.path.exists(barcodes_done):
                os.remove(barcodes_done)

            bc_threads = 1 if args.mode.enforces_single_thread() else args.threads
            bc_args = BarcodeCallingArgs(input_files, args.barcode_whitelist, args.mode,
                                         output_barcodes_list, output_fasta_list, sample.aux_dir, bc_threads,
                                         molecule=getattr(args, 'molecule', None))
            # Launching barcode calling in a separate process has the following reason:
            # Read chunks are not cleared by the GC in the end of barcode calling, leaving the main
            # IsoQuant process to consume ~2,5 GB even when barcode calling is done.
            # Once 16 child processes are created later, IsoQuant instantly takes threads x 2,5 GB for nothing.
            log_file, log_level = _get_log_params()
            with ProcessPoolExecutor(max_workers=1,
                                     initializer=setup_worker_logging,
                                     initargs=(log_file, log_level)) as proc:
                logger.info("Detecting barcodes for %d file(s)" % len(input_files))
                if bc_threads == 1:
                    future_res = proc.submit(process_single_thread, bc_args)
                else:
                    future_res = proc.submit(process_in_parallel, bc_args)

            concurrent.futures.wait([future_res],  return_when=concurrent.futures.ALL_COMPLETED)
            if future_res.exception() is not None:
                raise future_res.exception()

        # Mark all files as done and add to barcoded_reads
        for i, (input_file, output_barcodes, barcodes_done) in enumerate(zip(input_files, output_barcodes_list, barcodes_done_list)):
            sample.barcoded_reads.append(output_barcodes)
            open(barcodes_done, "w").close()
            logger.info("Processed %s, barcodes are stored in %s" % (input_file, output_barcodes))

        if args.mode.produces_new_fasta():
            logger.info("Reads were split during barcode calling")
            logger.info("The following files will be used instead of original reads %s " % ", ".join(map(lambda x: x[0], new_reads)))
            sample.file_list = new_reads


def run_pipeline(args):
    logger.info(" === IsoQuant pipeline started === ")
    logger.info("Python version: %s" % sys.version)
    logger.info("gffutils version: %s" % gffutils.__version__)
    logger.info("pysam version: %s" % pysam.__version__)
    logger.info("pyfaidx version: %s" % pyfaidx.__version__)
    if args.mode.needs_barcode_calling():
        # call barcodes
        call_barcodes(args)

    # gunzip refernece genome if needed
    prepare_reference_genome(args)

    # convert GTF/GFF if needed
    if args.genedb and not args.genedb.lower().endswith('db'):
        args.original_annotation = args.genedb
        args.genedb = convert_gtf_to_db(args)

    # map reads if fastqs are provided
    if args.input_data.input_type.needs_mapping():
        # substitute input reads with bams
        dataset_mapper = DataSetReadMapper(args)
        args.index = dataset_mapper.index_fname
        args.input_data = dataset_mapper.map_reads(args)

    if args.run_aligner_only:
        logger.info("Isoform assignment step is skipped because --run-aligner-only option was used")
        return

    # run isoform assignment
    dataset_processor = DatasetProcessor(args)
    dataset_processor.process_all_samples(args.input_data)

    # aggregate counts for all samples
    if len(args.input_data.samples) > 1 and args.genedb:
        combine_counts(args.input_data, args.output)

    logger.info(" === IsoQuant pipeline finished === ")


# Test mode is triggered by --test option
class TestMode(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        self.out_dir = 'isoquant_test'
        if os.path.exists(self.out_dir):
            shutil.rmtree(self.out_dir)
        source_dir = os.path.dirname(os.path.realpath(__file__))
        test_data_dir = os.path.join(source_dir, 'tests', 'simple_data')
        if not os.path.isdir(test_data_dir):
            # pip-installed: find test data via tests package location
            import tests
            test_data_dir = os.path.join(os.path.dirname(os.path.realpath(tests.__file__)), 'simple_data')
        if not os.path.isdir(test_data_dir):
            sys.stderr.write("ERROR: Test data not found. Cannot run in test mode.\n")
            sys.exit(1)
        options = ['--output', self.out_dir, '--threads', '2',
                   '--fastq', os.path.join(test_data_dir, 'chr9.4M.ont.sim.fq.gz'),
                   '--reference', os.path.join(test_data_dir, 'chr9.4M.fa.gz'),
                   '--genedb', os.path.join(test_data_dir, 'chr9.4M.gtf.gz'),
                   '--clean_start', '--data_type', 'nanopore', '--complete_genedb', '--force', '-p', 'TEST_DATA']
        print('=== Running in test mode === ')
        print("Running IsoQuant in test mode with the following options:")
        print(' '.join(options))
        print('Any other option is ignored ')
        main(options)
        if self._check_log():
            logger.info(' === TEST PASSED CORRECTLY === ')
        else:
            logger.error(' === TEST FAILED ===')
            sys.exit(IsoQuantExitCode.TEST_FAILED)
        parser.exit()


    def _check_log(self):
        with open(os.path.join(self.out_dir, 'isoquant.log'), 'r') as f:
            log = f.read()

        correct_results = ['total assignments 4', 'polyA tail detected in 2', 'unique: 1', 'known: 2', 'Processed 1 experiment']
        return all([result in log for result in correct_results])


def main(cmd_args):
    args, parser = parse_args(cmd_args)
    if not cmd_args:
        parser.print_usage()
        sys.exit(IsoQuantExitCode.SUCCESS)
    set_logger(args)
    args = check_and_load_args(args, parser)
    create_output_dirs(args)
    set_additional_params(args)
    run_pipeline(args)


def main_entry():
    """Entry point for console_scripts (pip install)."""
    try:
        main(sys.argv[1:])
    except SystemExit:
        raise
    except KeyboardInterrupt:
        raise
    except:
        if logger.handlers:
            strout = StringIO()
            print_exc(file=strout)
            s = strout.getvalue()
            if s:
                logger.critical("IsoQuant failed with the following error, please, submit this issue to "
                                "https://github.com/ablab/IsoQuant/issues\n" + s)
            else:
                print_exc()
        else:
            sys.stderr.write("IsoQuant failed with the following error, please, submit this issue to "
                             "https://github.com/ablab/IsoQuant/issues\n")
            print_exc()
        sys.exit(IsoQuantExitCode.UNCAUGHT_EXCEPTION)


if __name__ == "__main__":
    # stuff only to run when not called via 'import' here
    try:
        main(sys.argv[1:])
    except SystemExit:
        raise
    except KeyboardInterrupt:
        raise
    except:
        if logger.handlers:
            strout = StringIO()
            print_exc(file=strout)
            s = strout.getvalue()
            if s:
                logger.critical("IsoQuant failed with the following error, please, submit this issue to "
                                "https://github.com/ablab/IsoQuant/issues\n" + s)
            else:
                print_exc()
        else:
            sys.stderr.write("IsoQuant failed with the following error, please, submit this issue to "
                             "https://github.com/ablab/IsoQuant/issues\n")
            print_exc()
        sys.exit(IsoQuantExitCode.UNCAUGHT_EXCEPTION)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import unittest
import tempfile
import os
from src.stats import EnumStats
from enum import Enum


class TestEnum(Enum):
    value1 = 1
    value2 = 2
    value3 = 3


class TestEnumStats(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.test_file = os.path.join(self.temp_dir, "test_stats.txt")
        
    def tearDown(self):
        if os.path.exists(self.temp_dir):
            import shutil
            shutil.rmtree(self.temp_dir)
    
    def test_add_single_value(self):
        stats = EnumStats()
        stats.add(TestEnum.value1)
        self.assertEqual(stats.stats_dict[TestEnum.value1], 1)
        
    def test_add_multiple_values(self):
        stats = EnumStats()
        stats.add(TestEnum.value1)
        stats.add(TestEnum.value1)
        stats.add(TestEnum.value2)
        self.assertEqual(stats.stats_dict[TestEnum.value1], 2)
        self.assertEqual(stats.stats_dict[TestEnum.value2], 1)
        
    def test_add_with_count(self):
        stats = EnumStats()
        stats.add(TestEnum.value1, 5)
        self.assertEqual(stats.stats_dict[TestEnum.value1], 5)
        
    def test_merge(self):
        stats1 = EnumStats()
        stats1.add(TestEnum.value1, 3)
        stats1.add(TestEnum.value2, 2)
        
        stats2 = EnumStats()
        stats2.add(TestEnum.value1, 2)
        stats2.add(TestEnum.value3, 1)
        
        stats1.merge(stats2)
        self.assertEqual(stats1.stats_dict[TestEnum.value1], 5)
        self.assertEqual(stats1.stats_dict[TestEnum.value2], 2)
        self.assertEqual(stats1.stats_dict[TestEnum.value3], 1)
        
    def test_dump_and_load(self):
        stats = EnumStats()
        stats.add(TestEnum.value1, 3)
        stats.add(TestEnum.value2, 5)
        
        stats.dump(self.test_file)
        self.assertTrue(os.path.exists(self.test_file))
        
        loaded_stats = EnumStats(self.test_file)
        self.assertEqual(loaded_stats.stats_dict[TestEnum.value1], 3)
        self.assertEqual(loaded_stats.stats_dict[TestEnum.value2], 5)



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Stereo-seq barcode detectors.

Stereo-seq uses spatial barcodes with linker sequences and optional TSO detection.
"""

import logging
from typing import List, Optional

from ..indexers import ArrayKmerIndexer, Dict2BitKmerIndexer, SharedMemoryArray2BitKmerIndexer
from ..common import (
    find_polyt_start, reverese_complement, batch_str_to_2bit_chunked,
    find_candidate_with_max_score_ssw, detect_exact_positions, detect_first_exact_positions,
)
from .base import LinkerBarcodeDetectionResult, TSOBarcodeDetectionResult, SplittingBarcodeDetectionResult

logger = logging.getLogger('IsoQuant')


class StereoBarcodeDetector:
    """Stereo-seq barcode detector for standard (non-splitting) mode."""

    LINKER = "TTGTCTTCCTAAGAC"
    TSO_PRIMER = "ACTGAGAGGCATGGCGACCTTATCAG"
    PC1_PRIMER = "CTTCCGATCTATGGCGACCTTATCAG"
    BC_LENGTH = 25
    UMI_LEN = 10
    NON_T_UMI_BASES = 0
    UMI_LEN_DELTA = 4
    TERMINAL_MATCH_DELTA = 3
    STRICT_TERMINAL_MATCH_DELTA = 1

    def __init__(self, barcodes: List[str]):
        """
        Initialize Stereo-seq detector.

        Args:
            barcodes: List of known barcode sequences
        """
        self.main_primer = StereoBarcodeDetector.PC1_PRIMER
        self.pcr_primer_indexer = ArrayKmerIndexer([self.main_primer], kmer_size=6)
        self.linker_indexer = ArrayKmerIndexer([StereoBarcodeDetector.LINKER], kmer_size=5)
        self.strict_linker_indexer = ArrayKmerIndexer([StereoBarcodeDetector.LINKER], kmer_size=6)

        self.barcode_indexer = None
        if barcodes:
            bit_barcodes = batch_str_to_2bit_chunked(iter(barcodes), seq_len=self.BC_LENGTH)
            self.barcode_indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
            logger.info("Indexed %d barcodes" % len(bit_barcodes))

        self.min_score = 22

    def find_barcode_umi_multiple(self, read_id: str, sequence: str) -> List[LinkerBarcodeDetectionResult]:
        """Find multiple barcodes in a single read."""
        read_result = []
        r = self._find_barcode_umi_fwd(read_id, sequence)
        current_start = 0
        while r.polyT != -1:
            r.set_strand("+")
            read_result.append(r)
            new_start = r.polyT + 50
            current_start += new_start
            if len(sequence) - current_start < 50:
                break
            seq = sequence[current_start:]
            new_id = read_id + "_%d" % current_start
            r = self._find_barcode_umi_fwd(new_id, seq)

        rev_seq = reverese_complement(sequence)
        read_id += "_R"
        rr = self._find_barcode_umi_fwd(read_id, rev_seq)
        current_start = 0
        while rr.polyT != -1:
            rr.set_strand("-")
            read_result.append(rr)
            new_start = rr.polyT + 50
            current_start += new_start
            if len(rev_seq) - current_start < 50:
                break
            seq = rev_seq[current_start:]
            new_id = read_id + "_%d" % current_start
            rr = self._find_barcode_umi_fwd(new_id, seq)

        if not read_result:
            read_result.append(r)
        return read_result

    def find_barcode_umi(self, read_id: str, sequence: str) -> LinkerBarcodeDetectionResult:
        """Detect barcode and UMI in a read sequence."""
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    def _find_barcode_umi_fwd(self, read_id: str, sequence: str) -> LinkerBarcodeDetectionResult:
        polyt_start = find_polyt_start(sequence)

        linker_start, linker_end = None, None
        if polyt_start != -1:
            # use relaxed parameters if polyA is found
            linker_occurrences = self.linker_indexer.get_occurrences(sequence[0:polyt_start + 1])
            linker_start, linker_end = detect_exact_positions(sequence, 0, polyt_start + 1,
                                                              self.linker_indexer.k, StereoBarcodeDetector.LINKER,
                                                              linker_occurrences, min_score=12,
                                                              start_delta=self.TERMINAL_MATCH_DELTA,
                                                              end_delta=self.TERMINAL_MATCH_DELTA)

        if linker_start is None:
            # if polyT was not found, or linker was not found to the left of polyT,
            # look for linker in the entire read
            linker_occurrences = self.strict_linker_indexer.get_occurrences(sequence)
            linker_start, linker_end = detect_exact_positions(sequence, 0, len(sequence),
                                                              self.linker_indexer.k, StereoBarcodeDetector.LINKER,
                                                              linker_occurrences, min_score=12,
                                                              start_delta=self.STRICT_TERMINAL_MATCH_DELTA,
                                                              end_delta=self.STRICT_TERMINAL_MATCH_DELTA)

        if linker_start is None:
            return LinkerBarcodeDetectionResult(read_id, polyT=polyt_start)
        logger.debug("LINKER: %d-%d" % (linker_start, linker_end))

        if polyt_start == -1:
            # if polyT was not detected earlier, use relaxed parameters once the linker is found
            presumable_polyt_start = linker_end + self.UMI_LEN
            search_start = presumable_polyt_start - 4
            search_end = min(len(sequence), presumable_polyt_start + 10)
            polyt_start = find_polyt_start(sequence[search_start:search_end], window_size=5, polya_fraction=1.0)
            if polyt_start != -1:
                polyt_start += search_start

        primer_occurrences = self.pcr_primer_indexer.get_occurrences(sequence[:linker_start])
        primer_start, primer_end = detect_exact_positions(sequence, 0, linker_start,
                                                          self.pcr_primer_indexer.k, self.main_primer,
                                                          primer_occurrences, min_score=12,
                                                          end_delta=self.TERMINAL_MATCH_DELTA)
        if primer_start is not None:
            logger.debug("PRIMER: %d-%d" % (primer_start, primer_end))
        else:
            primer_start = -1
            primer_end = linker_start - self.BC_LENGTH - 1

        if primer_end < 0:
            return LinkerBarcodeDetectionResult(read_id, polyT=polyt_start, primer=-1,
                                                linker_start=linker_start, linker_end=linker_end)

        barcode_start = primer_end + 1
        barcode_end = linker_start - 1
        bc_len = barcode_end - barcode_start
        if abs(bc_len - self.BC_LENGTH) > 10:
            return LinkerBarcodeDetectionResult(read_id, polyT=polyt_start, primer=primer_end,
                                                linker_start=linker_start, linker_end=linker_end)

        potential_barcode = sequence[barcode_start:barcode_end + 1]
        logger.debug("Barcode: %s" % (potential_barcode))
        matching_barcodes = self.barcode_indexer.get_occurrences(potential_barcode, max_hits=10, min_kmers=2)
        barcode, bc_score, bc_start, bc_end = \
            find_candidate_with_max_score_ssw(matching_barcodes, potential_barcode,
                                              min_score=self.min_score, sufficient_score=self.BC_LENGTH - 1)

        if barcode is None:
            return LinkerBarcodeDetectionResult(read_id, polyT=polyt_start, primer=primer_end,
                                                linker_start=linker_start, linker_end=linker_end)
        logger.debug("Found: %s %d-%d" % (barcode, bc_start, bc_end))

        potential_umi_start = linker_end + 1
        potential_umi_end = polyt_start - 1
        if potential_umi_end - potential_umi_start <= 5:
            potential_umi_end = potential_umi_start + self.UMI_LEN - 1
        umi = None
        good_umi = False
        if potential_umi_start + 2 * self.UMI_LEN > potential_umi_end > potential_umi_start:
            umi = sequence[potential_umi_start:potential_umi_end + 1]
            logger.debug("Potential UMI: %s" % umi)
            good_umi = abs(len(umi) - self.UMI_LEN) <= self.UMI_LEN_DELTA

        if not umi:
            return LinkerBarcodeDetectionResult(read_id, barcode, BC_score=bc_score,
                                                polyT=polyt_start, primer=primer_end,
                                                linker_start=linker_start, linker_end=linker_end)
        return LinkerBarcodeDetectionResult(read_id, barcode, umi, bc_score, good_umi,
                                            polyT=polyt_start, primer=primer_end,
                                            linker_start=linker_start, linker_end=linker_end)

    @staticmethod
    def result_type():
        return LinkerBarcodeDetectionResult

    @classmethod
    def header(cls):
        return cls.result_type().header()


class SharedMemoryStereoBarcodeDetector(StereoBarcodeDetector):
    """Stereo-seq detector with shared memory for large barcode sets."""

    MIN_BARCODES_FOR_SHARED_MEM = 1000000

    def __init__(self, barcodes: List[str]):
        super().__init__([])
        # Convert to 2-bit encoding first (memory efficient)
        bit_barcodes = batch_str_to_2bit_chunked(iter(barcodes), seq_len=self.BC_LENGTH)
        self.barcode_count = len(bit_barcodes)
        self.bit_barcodes = None
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            # For small/medium sets, use Dict2BitKmerIndexer
            self.bit_barcodes = bit_barcodes
            self.barcode_indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
        else:
            # For large sets, use shared memory indexer
            self.barcode_indexer = SharedMemoryArray2BitKmerIndexer(bit_barcodes, kmer_size=14,
                                                                    seq_len=self.BC_LENGTH)

        logger.info("Indexed %d barcodes" % self.barcode_count)

    def __getstate__(self):
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            return (self.min_score,
                    self.barcode_count,
                    self.bit_barcodes)
        else:
            return (self.min_score,
                    self.barcode_count,
                    self.barcode_indexer.get_sharable_info())

    def __setstate__(self, state):
        self.min_score = state[0]
        super().__init__([])
        self.bit_barcodes = None
        self.barcode_count = state[1]
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            self.bit_barcodes = state[2]
            self.barcode_indexer = Dict2BitKmerIndexer(self.bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
        else:
            self.barcode_indexer = SharedMemoryArray2BitKmerIndexer.from_sharable_info(state[2])


class StereoSplittingBarcodeDetector:
    """Stereo-seq detector with read splitting for concatenated reads."""

    TSO5 = "CCCGCCTCTCAGTACGTCAGCAG"
    LINKER = "TTGTCTTCCTAAGAC"
    TSO_PRIMER = "ACTGAGAGGCATGGCGACCTTATCAG"
    PC1_PRIMER = "CTTCCGATCTATGGCGACCTTATCAG"
    BC_LENGTH = 25
    UMI_LEN = 10
    NON_T_UMI_BASES = 0
    UMI_LEN_DELTA = 3
    TERMINAL_MATCH_DELTA = 3
    STRICT_TERMINAL_MATCH_DELTA = 1

    def __init__(self, barcodes: List[str]):
        """
        Initialize splitting detector.

        Args:
            barcodes: List of known barcode sequences
        """
        self.main_primer = self.PC1_PRIMER
        self.tso5_indexer = ArrayKmerIndexer([self.TSO5], kmer_size=8)
        self.pcr_primer_indexer = ArrayKmerIndexer([self.main_primer], kmer_size=6)
        self.linker_indexer = ArrayKmerIndexer([self.LINKER], kmer_size=5)
        self.strict_linker_indexer = ArrayKmerIndexer([StereoBarcodeDetector.LINKER], kmer_size=7)

        self.barcode_indexer = None
        if barcodes:
            bit_barcodes = batch_str_to_2bit_chunked(iter(barcodes), seq_len=self.BC_LENGTH)
            self.barcode_indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
            logger.info("Indexed %d barcodes" % len(bit_barcodes))
        self.min_score = 22

    def find_barcode_umi(self, read_id: str, sequence: str) -> SplittingBarcodeDetectionResult:
        """Find multiple barcodes in a concatenated read."""
        read_result = SplittingBarcodeDetectionResult(read_id)
        logger.debug("Looking in forward direction")
        r = self._find_barcode_umi_fwd(read_id, sequence)
        prev_start = 0
        while r.polyT != -1:
            r.set_strand("+")
            read_result.append(r)
            if r.tso5 != -1:
                current_start = r.tso5 + 15
            else:
                current_start = r.polyT + 100
            # always make a step
            current_start = max(prev_start + 150, current_start)
            prev_start = current_start
            if len(sequence) - current_start < 50:
                break

            logger.debug("Looking further from %d" % current_start)
            seq = sequence[current_start:]
            r = self._find_barcode_umi_fwd(read_id, seq)
            r.update_coordinates(current_start)

        logger.debug("Looking in reverse direction")
        rev_seq = reverese_complement(sequence)
        r = self._find_barcode_umi_fwd(read_id, rev_seq)
        prev_start = 0
        while r.polyT != -1:
            r.set_strand("-")
            read_result.append(r)
            if r.tso5 != -1:
                current_start = r.tso5 + 15
            else:
                current_start = r.polyT + 100
            # always make a step
            current_start = max(prev_start + 150, current_start)
            prev_start = current_start
            if len(rev_seq) - current_start < 50:
                break

            logger.debug("Looking further from %d" % current_start)
            seq = rev_seq[current_start:]
            r = self._find_barcode_umi_fwd(read_id, seq)
            r.update_coordinates(current_start)

        if read_result.empty():
            # add empty result anyway
            read_result.append(r)

        read_result.filter()
        logger.debug("Total barcodes detected %d" % len(read_result.detected_patterns))
        return read_result

    def find_barcode_umi_single(self, read_id: str, sequence: str) -> TSOBarcodeDetectionResult:
        """Find single barcode (non-splitting mode)."""
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    def _find_barcode_umi_fwd(self, read_id: str, sequence: str) -> TSOBarcodeDetectionResult:
        polyt_start = find_polyt_start(sequence)
        logger.debug("PolyT found right away %d" % polyt_start)

        linker_start, linker_end = None, None
        tso5_start = None
        if polyt_start != -1:
            # use relaxed parameters if polyA is found
            logger.debug("Looking for linker in %d" % len(sequence[0:polyt_start + 1]))
            linker_occurrences = self.linker_indexer.get_occurrences(sequence[0:polyt_start + 1])
            linker_start, linker_end = detect_exact_positions(sequence, 0, polyt_start + 1,
                                                              self.linker_indexer.k, self.LINKER,
                                                              linker_occurrences, min_score=10,
                                                              start_delta=self.TERMINAL_MATCH_DELTA,
                                                              end_delta=self.TERMINAL_MATCH_DELTA)

            tso5_occurrences = self.tso5_indexer.get_occurrences(sequence[polyt_start + 1:])
            tso5_start, tso5_end = detect_first_exact_positions(sequence, polyt_start + 1, len(sequence),
                                                                self.tso5_indexer.k, self.TSO5,
                                                                tso5_occurrences, min_score=15,
                                                                start_delta=self.TERMINAL_MATCH_DELTA,
                                                                end_delta=self.TERMINAL_MATCH_DELTA)

        if linker_start is None:
            # if polyT was not found, or linker was not found to the left of polyT,
            # look for linker in the entire read
            linker_occurrences = self.strict_linker_indexer.get_occurrences(sequence)
            linker_start, linker_end = detect_first_exact_positions(sequence, 0, len(sequence),
                                                                    self.linker_indexer.k, StereoBarcodeDetector.LINKER,
                                                                    linker_occurrences, min_score=12,
                                                                    start_delta=self.STRICT_TERMINAL_MATCH_DELTA,
                                                                    end_delta=self.STRICT_TERMINAL_MATCH_DELTA)

        if linker_start is None:
            return TSOBarcodeDetectionResult(read_id, polyT=polyt_start)
        logger.debug("LINKER: %d-%d" % (linker_start, linker_end))

        if polyt_start == -1 or polyt_start < linker_start:
            # if polyT was not detected earlier, use relaxed parameters once the linker is found
            presumable_polyt_start = linker_end + self.UMI_LEN
            search_start = presumable_polyt_start - 4
            search_end = min(len(sequence), presumable_polyt_start + 10)
            polyt_start = find_polyt_start(sequence[search_start:search_end], window_size=5, polya_fraction=1.0)
            if polyt_start != -1:
                polyt_start += search_start
                logger.debug("PolyT found later %d" % polyt_start)
            else:
                logger.debug("PolyT was not found %d" % polyt_start)

            tso5_occurrences = self.tso5_indexer.get_occurrences(sequence[polyt_start + 1:])
            tso5_start, tso5_end = detect_first_exact_positions(sequence, polyt_start + 1, len(sequence),
                                                                self.tso5_indexer.k, self.TSO5,
                                                                tso5_occurrences, min_score=15,
                                                                start_delta=self.TERMINAL_MATCH_DELTA,
                                                                end_delta=self.TERMINAL_MATCH_DELTA)

        if tso5_start is not None:
            logger.debug("TSO found %d" % tso5_start)
            # check that no another linker is found inbetween polyA and TSO 5'
            linker_occurrences = self.strict_linker_indexer.get_occurrences(sequence[polyt_start + 1: tso5_start])
            new_linker_start, new_linker_end = detect_exact_positions(sequence, polyt_start + 1, tso5_start,
                                                                      self.linker_indexer.k, self.LINKER,
                                                                      linker_occurrences, min_score=12,
                                                                      start_delta=self.STRICT_TERMINAL_MATCH_DELTA,
                                                                      end_delta=self.STRICT_TERMINAL_MATCH_DELTA)

            if new_linker_start is not None and new_linker_start != -1 and new_linker_start - polyt_start > 100:
                # another linker found inbetween polyT and TSO
                logger.debug("Another linker was found before TSO: %d" % new_linker_start)
                tso5_start = new_linker_start - self.BC_LENGTH - len(self.main_primer) - len(self.TSO5)
                logger.debug("TSO updated %d" % tso5_start)
        else:
            tso5_start = -1

        primer_occurrences = self.pcr_primer_indexer.get_occurrences(sequence[:linker_start])
        primer_start, primer_end = detect_exact_positions(sequence, 0, linker_start,
                                                          self.pcr_primer_indexer.k, self.main_primer,
                                                          primer_occurrences, min_score=12,
                                                          end_delta=self.TERMINAL_MATCH_DELTA)
        if primer_start is not None:
            logger.debug("PRIMER: %d-%d" % (primer_start, primer_end))
        else:
            primer_end = linker_start - self.BC_LENGTH - 1

        if primer_end < 0:
            return TSOBarcodeDetectionResult(read_id, polyT=polyt_start, primer=-1,
                                             linker_start=linker_start, linker_end=linker_end, tso=tso5_start)

        barcode_start = primer_end + 1
        barcode_end = linker_start - 1
        bc_len = barcode_end - barcode_start
        if abs(bc_len - self.BC_LENGTH) > 10:
            return TSOBarcodeDetectionResult(read_id, polyT=polyt_start, primer=primer_end,
                                             linker_start=linker_start, linker_end=linker_end, tso=tso5_start)

        potential_barcode = sequence[barcode_start:barcode_end + 1]
        logger.debug("Barcode: %s" % (potential_barcode))
        if not self.barcode_indexer:
            return TSOBarcodeDetectionResult(read_id, potential_barcode, BC_score=0,
                                             polyT=polyt_start, primer=primer_end,
                                             linker_start=linker_start, linker_end=linker_end, tso=tso5_start)

        matching_barcodes = self.barcode_indexer.get_occurrences(potential_barcode, max_hits=10, min_kmers=2)
        barcode, bc_score, bc_start, bc_end = \
            find_candidate_with_max_score_ssw(matching_barcodes, potential_barcode,
                                              min_score=self.min_score, sufficient_score=self.BC_LENGTH - 1)

        if barcode is None:
            return TSOBarcodeDetectionResult(read_id, polyT=polyt_start, primer=primer_end,
                                             linker_start=linker_start, linker_end=linker_end, tso=tso5_start)
        logger.debug("Found: %s %d-%d" % (barcode, bc_start, bc_end))

        potential_umi_start = linker_end + 1
        potential_umi_end = polyt_start - 1
        if potential_umi_end - potential_umi_start <= 5:
            potential_umi_end = potential_umi_start + self.UMI_LEN - 1
        umi = None
        good_umi = False
        if potential_umi_start + 2 * self.UMI_LEN > potential_umi_end > potential_umi_start:
            umi = sequence[potential_umi_start:potential_umi_end + 1]
            logger.debug("Potential UMI: %s" % umi)
            good_umi = abs(len(umi) - self.UMI_LEN) <= self.UMI_LEN_DELTA

        if not umi:
            return TSOBarcodeDetectionResult(read_id, barcode, BC_score=bc_score,
                                             polyT=polyt_start, primer=primer_end,
                                             linker_start=linker_start, linker_end=linker_end, tso=tso5_start)
        return TSOBarcodeDetectionResult(read_id, barcode, umi, bc_score, good_umi,
                                         polyT=polyt_start, primer=primer_end,
                                         linker_start=linker_start, linker_end=linker_end, tso=tso5_start)

    @staticmethod
    def result_type():
        return SplittingBarcodeDetectionResult

    @classmethod
    def header(cls):
        return cls.result_type().header()


class SharedMemoryStereoSplittingBarcodeDetector(StereoSplittingBarcodeDetector):
    """Stereo-seq splitting detector with shared memory for large barcode sets."""

    MIN_BARCODES_FOR_SHARED_MEM = 1000000

    def __init__(self, barcodes: List[str]):
        super().__init__([])
        # Convert to 2-bit encoding first (memory efficient)
        bit_barcodes = batch_str_to_2bit_chunked(iter(barcodes), seq_len=self.BC_LENGTH)
        self.barcode_count = len(bit_barcodes)
        self.bit_barcodes = None
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            # For small/medium sets, use Dict2BitKmerIndexer
            self.bit_barcodes = bit_barcodes
            self.barcode_indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
        else:
            # For large sets, use shared memory indexer
            self.barcode_indexer = SharedMemoryArray2BitKmerIndexer(bit_barcodes, kmer_size=14,
                                                                    seq_len=self.BC_LENGTH)

        logger.info("Indexed %d barcodes" % self.barcode_count)

    def __getstate__(self):
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            return (self.min_score,
                    self.barcode_count,
                    self.bit_barcodes)
        else:
            return (self.min_score,
                    self.barcode_count,
                    self.barcode_indexer.get_sharable_info())

    def __setstate__(self, state):
        self.min_score = state[0]
        super().__init__([])
        self.bit_barcodes = None
        self.barcode_count = state[1]
        if self.barcode_count < self.MIN_BARCODES_FOR_SHARED_MEM:
            self.bit_barcodes = state[2]
            self.barcode_indexer = Dict2BitKmerIndexer(self.bit_barcodes, kmer_size=14, seq_len=self.BC_LENGTH)
        else:
            self.barcode_indexer = SharedMemoryArray2BitKmerIndexer.from_sharable_info(state[2])


class SharedMemoryWrapper:
    """Generic wrapper to add shared memory support to any barcode detector."""

    def __init__(self, barcode_detector_class, barcodes: List[str]):
        """
        Initialize wrapper.

        Args:
            barcode_detector_class: Detector class to wrap
            barcodes: List of known barcodes
        """
        self.barcode_detector_class = barcode_detector_class
        self.barcode_detector = self.barcode_detector_class([])
        self.min_score = self.barcode_detector.min_score
        # Convert to 2-bit and use shared memory indexer
        barcode_iter = iter(barcodes) if isinstance(barcodes, list) else barcodes
        bit_barcodes = batch_str_to_2bit_chunked(barcode_iter, seq_len=self.barcode_detector_class.BC_LENGTH)
        self.barcode_detector.barcode_indexer = SharedMemoryArray2BitKmerIndexer(bit_barcodes, kmer_size=14,
                                                                                 seq_len=self.barcode_detector_class.BC_LENGTH)

        logger.info("Indexed %d barcodes" % self.barcode_detector.barcode_indexer.total_sequences)

    def __getstate__(self):
        return (self.barcode_detector_class,
                self.min_score,
                self.barcode_detector.barcode_indexer.get_sharable_info())

    def __setstate__(self, state):
        self.barcode_detector_class = state[0]
        self.min_score = state[1]
        self.barcode_detector = self.barcode_detector_class([])
        self.barcode_detector.min_score = self.min_score
        self.barcode_detector.barcode_indexer = SharedMemoryArray2BitKmerIndexer.from_sharable_info(state[2])



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
import unittest

import src.common as c


class TestMiscFunction:
    def test_get_first(self):
        assert c.get_first_best_from_sorted([(1,1), (2,2), (3,3)]) == [1]
        assert c.get_first_best_from_sorted([(1,1), (2,1), (3,3)]) == [1, 2]

    @pytest.mark.parametrize("collection, elem, expected",
                             [([0, 1, 2], 2, 2), ([0, 1, 2], 0, 0), ([0, 1, 2], 1, 1), ([1, 1, 0], 1, 1),
                              ([1, 1, 1], 1, 2)],
                             ids=("last", "first", "middle", "several", "all"))
    def test_correct(self, collection, elem, expected):
        assert c.rindex(collection, elem) == expected

    @pytest.mark.parametrize("collection", [(), ("b", "c")])
    def test_no_such_elem(self, collection):
        with pytest.raises(ValueError, match="a is not in list"):
            c.rindex(collection, "a")


class TestRanges:
    @pytest.mark.parametrize("range1, range2, delta, expected",
                             [((1, 20), (1, 20), 0, True), ((1, 20), (2, 21), 1, True), ((1, 10), (1, 20), 5, False)],
                             ids=("equals", "less_than_delta", "bigger_than_delta"))
    def test_equal(self, range1, range2, delta, expected):
        assert c.equal_ranges(range1, range2, delta) == expected

    @pytest.mark.parametrize("range1, range2, expected",
                             [((1, 20), (1, 20), True), ((1, 20), (20, 121), True), ((1, 10), (11, 20), False),
                              ((111, 210), (11, 110), False), ((120, 210), (101, 121), True)])
    def test_overlaps(self, range1, range2, expected):
        assert c.overlaps(range1, range2) == expected

    @pytest.mark.parametrize("range1, range2, delta, expected",
                             [((1, 20), (20, 25), 0, True), ((18, 120), (2, 20), 3, True), ((1, 10), (6, 20), 6, False)])
    def test_overlaps_at_least(self, range1, range2, delta, expected):
        assert c.overlaps_at_least(range1, range2, delta) == expected

    @pytest.mark.parametrize("range1, range2, delta, expected",
                             [((1, 25), (20, 25), 0, True), ((18, 40), (16, 41), 3, True), ((10, 27), (5, 30), 4, False)])
    def test_contains_approx(self, range1, range2, delta, expected):
        assert c.contains_approx(range1, range2, delta) == expected

    @pytest.mark.parametrize("range1, range2, expected",
                             [((1, 20), (1, 20), True), ((1, 120), (20, 101), True), ((1, 10), (11, 20), False),
                              ((111, 210), (111, 211), False), ((120, 210), (119, 121), False)])
    def test_contains(self, range1, range2, expected):
        assert c.contains(range1, range2) == expected

    @pytest.mark.parametrize("range1, range2, expected",
                             [((1, 20), (1, 20), True), ((1, 120), (20, 201), False), ((1, 10), (11, 20), False),
                              ((111, 210), (111, 211), False), ((120, 210), (119, 201), True)])
    def test_covers_start(self, range1, range2, expected):
        assert c.covers_start(range1, range2) == expected

    @pytest.mark.parametrize("range1, range2, expected",
                             [((1, 20), (1, 20), True), ((1, 120), (20, 201), True), ((1, 10), (11, 20), False),
                              ((111, 210), (111, 211), True), ((120, 210), (121, 211), True)])
    def test_covers_end(self, range1, range2, expected):
        assert c.covers_end(range1, range2) == expected

    @pytest.mark.parametrize("range, expected",
                             [((1, 20), 20), ((1, 1), 1)])
    def test_interval_len(self, range, expected):
        assert c.interval_len(range) == expected

    @pytest.mark.parametrize("ranges, expected",
                             [([(1, 20), (40, 50)], 31), ([(1, 1), (5, 5)], 2)])
    def test_intervals_len(self, ranges, expected):
        assert c.intervals_total_length(ranges) == expected

    @pytest.mark.parametrize("ranges, point, expected",
                             [([(1, 20), (40, 50)], 60, 31), ([(1, 1), (5, 5)], 10, 2),
                              ([(1, 20), (40, 50)], 39, 20), ([(1, 20), (40, 50)], 41, 21)])
    def test_intervals_len_to_point(self, ranges, point, expected):
        assert c.sum_intervals_to_point(ranges, point) == expected

    @pytest.mark.parametrize("ranges, point, expected",
                             [([(1, 20), (40, 50)], 1, 30), ([(1, 1), (5, 5)], 0, 2),
                              ([(1, 20), (40, 50)], 39, 11), ([(1, 20), (40, 50)], 11, 20)])
    def test_intervals_len_from_point(self, ranges, point, expected):
        assert c.sum_intervals_from_point(ranges, point) == expected


class TestSimilarityFunctions:
    @pytest.mark.parametrize("ranges1, ranges2, expected",
                             [([(1, 70)], [(31, 100)], 0.4),
                              ([(1, 20), (41, 50), (61, 70)], [(1, 20), (41, 50), (61, 70)], 1.0),
                              ([(1, 20), (41, 50), (61, 70)], [(41, 50), (61, 70), (101, 160)], 0.2),
                              ([(1, 20), (41, 50), (61, 70)], [(46, 55), (60, 62), (70, 74), (101, 150)], 0.08),
                              ([(1, 20), (41, 50), (61, 70)], [(146, 155), (160, 172)], 0.0),
                              ([(146, 155), (160, 172)], [(1, 20), (41, 50), (61, 70)], 0.0),
                              ([(11, 20), (41, 50), (61, 70)], [(1, 4), (10, 12), (14, 16), (19, 30), (30, 34), (60, 70), (101, 149)], 0.17)])
    def test_jaccard_index(self, ranges1, ranges2, expected):
        assert c.jaccard_similarity(ranges1, ranges2) == expected

    @pytest.mark.parametrize("read_exons, isoform_exons, expected",
                             [([(1, 100)], [(31, 110)], 0.7),
                              ([(1, 20), (41, 50), (61, 70)], [(1, 20), (41, 50), (61, 70)], 1.0),
                              ([(1, 20), (41, 50), (61, 80)], [(41, 50), (61, 70), (101, 110)], 0.4),
                              ([(1, 20), (41, 50), (61, 80)], [(0, 0), (46, 55), (60, 62), (70, 74), (80,82), (101, 110)], 0.26),
                              ([(1, 20), (41, 50), (61, 80)], [(146, 155), (160, 172)], 0.0),
                              ([(146, 155), (160, 172)], [(1, 20), (41, 50), (61, 70)], 0.0),
                              ([(11, 20), (41, 50), (61, 70), (201, 220)], [(1, 4), (10, 12), (14, 16), (19, 30), (30, 34), (60, 70), (101, 109)], 0.34)])
    def test_read_coverage_fraction(self, read_exons, isoform_exons, expected):
        assert c.read_coverage_fraction(read_exons, isoform_exons) == expected

    @pytest.mark.parametrize("read_exons, isoform_region, expected",
                             [([(1, 100)], (31, 110), 0.3),
                              ([(1, 100)], (31, 70), 0.6),
                              ([(1, 20), (41, 50), (61, 70)], (1, 70), 0.0),
                              ([(1, 20), (41, 50), (61, 70)], (0, 71), 0.0),
                              ([(1, 20), (41, 50), (61, 80)], (35, 55), 0.8),
                              ([(1, 20), (41, 50), (61, 80)], (146, 172), 1.0),
                              ([(11, 20), (41, 50), (61, 90)], (16, 70), 0.5)])
    def test_extra_flanking(self, read_exons, isoform_region, expected):
        assert c.extra_exon_percentage(isoform_region, read_exons) == expected


class TestExonFunctions:
    @pytest.mark.parametrize("exons, expected",
                             [([(1, 100)], []),
                              ([(1, 20), (41, 50), (61, 70)], [(21, 40), (51, 60)])])
    def test_junctions_from_blocks(self, exons, expected):
        assert c.junctions_from_blocks(exons) == expected

    @pytest.mark.parametrize("region, introns, intron_position, expected",
                             [((0, 100), [(20, 80)], 0, (81, 100)),
                              ((0, 100), [(20, 40), (60, 80)], 0, (41, 59)),
                              ((0, 100), [(20, 40), (60, 80)], 1, (81, 100))])
    def test_following_exon(self, region, introns, intron_position, expected):
        assert c.get_following_exon_from_junctions(region, introns, intron_position) == expected

    @pytest.mark.parametrize("region, introns, intron_position, expected",
                             [((0, 100), [(20, 80)], 0, (0, 19)),
                              ((0, 100), [(20, 40), (60, 80)], 0, (0, 19)),
                              ((0, 100), [(20, 40), (60, 80)], 1, (41, 59))])
    def test_preceding_exon(self, region, introns, intron_position, expected):
        assert c.get_preceding_exon_from_junctions(region, introns, intron_position) == expected

    @pytest.mark.parametrize("region, introns, exon_position, expected",
                             [((0, 100), [(20, 80)], 0, (0, 19)),
                              ((0, 100), [(20, 80)], 1, (81, 100)),
                              ((0, 100), [(20, 40), (60, 80)], 0, (0, 19)),
                              ((0, 100), [(20, 40), (60, 80)], 1, (41, 59)),
                              ((0, 100), [(20, 40), (60, 80)], 2, (81, 100))])
    def test_get_exon(self, region, introns, exon_position, expected):
        assert c.get_exon(region, introns, exon_position) == expected

    @pytest.mark.parametrize("exons, expected",
                             [([(1, 100)], [(2, 100)]),
                              ([(1, 20), (41, 50), (61, 70)], [(2, 20), (42, 50), (62, 70)])])
    def test_correct_bam_coords(self, exons, expected):
        assert c.correct_bam_coords(exons) == expected

    @pytest.mark.parametrize("exons, cigar_tuples, expected",
                             [([(1, 100)], [(0, 99)], [(1, 100)]),
                              ([(10, 20), (20, 30), (41, 50), (51, 60), (61, 70), (72, 80), (100, 110), (110, 114), (117, 120)],
                               [(4, 10), (0, 10), (1, 2), (0, 10), (3, 11), (0, 9), (2, 1), (0, 9), (2, 1), (0, 9), (2, 2), (0, 8), (3, 20), (0, 10), (1, 3), (0, 4),  (2, 3), (0, 3), (5, 10)],
                               [(10, 30), (41, 80), (100, 120)])])
    def test_concat(self, exons, cigar_tuples, expected):
        assert c.concat_gapless_blocks(exons, cigar_tuples) == expected

    @pytest.mark.parametrize("ref_start, cigar_tuples, expected",
                             [(3, [(0, 4), (1, 1), (2, 1), (0, 2)], ([(4, 10)], [(0, 6)], [(0, 3)])),
                              (5, [(0, 2), (1, 1), (0, 5), (3, 3), (0, 3)],
                               ([(6, 12), (16, 18)], [(0, 7), (8, 10)], [(0, 2), (4, 4)])),
                              (3, [(0, 5), (2, 3), (0, 7)], ([(4, 18)], [(0, 11)], [(0, 2)])),
                              (2, [(0, 3), (2, 1), (0, 4), (3, 3), (1, 2), (0, 5)],
                               ([(3, 10), (14, 18)], [(0, 6), (7, 13)], [(0, 2), (4, 5)])),
                              (5, [(0, 3), (1, 1), (0, 4), (3, 3), (2, 2), (0, 5)],
                               ([(6, 12), (16, 22)], [(0, 7), (8, 12)], [(0, 2), (4, 5)])),
                              (6, [(4, 2), (0, 6), (2, 1), (0, 5), (1, 2), (0, 4)], ([(7, 22)], [(2, 18)], [(1, 5)])),
                              (1, [(4, 3), (0, 9), (3, 4), (0, 6), (2, 1), (0, 7), (4, 2)],
                               ([(2, 10), (15, 28)], [(3, 11), (12, 24)], [(1, 1), (3, 5)]))
                              ])
    def test_get_read_blocks(self, ref_start, cigar_tuples, expected):
        assert c.get_read_blocks(ref_start, cigar_tuples) == expected


class TestProfiles:
    @pytest.mark.parametrize("profile1, profile2, expected", [([-2, 1], [1, 1], True), ([-2, -1, -2], [-1, -1, -1], True),
                                                              ([1], [1], True),
                                                              ([-2, 1, -1, -2], [-1, 1, 1, 1], False),
                                                              ([-2, 1, -1, -2], [-1, 1, -1, 1], True)])
    def test_is_subprofile(self, profile1, profile2, expected):
        assert c.is_subprofile(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([], [], 0), ([0], [1], 0), ([1], [1], 0),
                                                              ([1, -1], [1, 1], 1),
                                                              ([0], [0], 0), ([0, 1], [1, 1], 0),
                                                              ([-1, 1], [1, 0], 1), ([1, 0, -1], [0, 1, 1], 1)])
    def test_difference_in_present_features(self, profile1, profile2, expected):
        assert c.difference_in_present_features(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, profile_range, expected",
                             [([], [], None, 0), ([0], [1], (0, 1), 0), ([1], [1], (0, 1), 0),
                              ([1, -1], [1, 1], (1, 2), 1),
                              ([0], [0], (0, 1), 0), ([0, 1], [1, 1], (0, 1), 0),
                              ([-1, 1], [1, 0], (0, 1), 1), ([1, 0, -1], [0, 1, 1], (2, 3), 1)])
    def test_difference_in_present_features_with_range(self, profile1, profile2, profile_range, expected):
        assert c.difference_in_present_features(profile1, profile2, profile_range=profile_range) == expected

    @pytest.mark.parametrize("profile1, profile2, diff_limit, expected",
                             [([1, -1, -1], [-1, 1, 1], 0, 1)])
    def test_difference_in_present_features_with_lim(self, profile1, profile2, diff_limit, expected):
        assert c.difference_in_present_features(profile1, profile2, diff_limit=diff_limit) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([0], [1], 0), ([0, 1], [1, 1], 1),
                                                              ([0, 1], [1, -2], 0), ([0, 1, -1], [-2, 1, 1], 1)])
    def test_count_both_present_features(self, profile1, profile2, expected):
        assert c.count_both_present_features(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], True), ([1, 1], [0, 1], False),
                                                              ([-1, 1, 1, -2], [0, 1, 1, -2], True),
                                                              ([-2, 1, 1], [0, 1, -1], False)])
    def test_all_features_present(self, profile1, profile2, expected):
        assert c.all_features_present(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], [1]), ([0, 1], [1, 1], [0, 1]),
                                                              ([0, 1, 1, 0], [-1, 1, 1, -2], [0, 1, 1, 0]),
                                                              ([0, 1, -1], [-2, 1, 1], [0, 1, 0])])
    def test_find_matching_positions(self, profile1, profile2, expected):
        assert c.find_matching_positions(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], True), ([0, 1], [1, 1], True),
                                                              ([0, 1, 1, 0], [-1, 1, 1, -2], True),
                                                              ([1, -1, 0, 0], [-1, 1, 1, -2], False),
                                                              ([0, 1, -1], [-2, 1, 1], True)])
    def test_has_overlapping_features(self, profile1, profile2, expected):
        assert c.has_overlapping_features(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], False), ([0, 1], [1, 1], False),
                                                              ([0, 1, 1, 0], [-1, 1, 1, -2], False),
                                                              ([1, -1, 0, 0], [-1, 1, 1, -2], True),
                                                              ([0, 1, -1], [-2, 1, 1], True)])
    def test_has_inconsistent_features(self, profile1, profile2, expected):
        assert c.has_inconsistent_features(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, mask, expected", [([1], [1], [1]), ([0, 1], [1, 1], [0, 1]),
                                                              ([0, 1, 1, 0], [1, 1, 0, 0], [0, 1, 0, 0]),
                                                              ([-2, -1, 1, 0], [1, 1, 0, 1], [-2, -1, 0, 0])])
    def test_mask_profile(self, profile1, mask, expected):
        assert c.mask_profile(profile1, mask) == expected

    @pytest.mark.parametrize("profile1, features, expected", [([1], [(10, 20)], [(10, 20)]), ([0, 1], [(10, 20), (30, 40)], [(30, 40)]),
                                                              ([0, 1, 0, 1], [(10, 20), (30, 40), (110, 120), (130, 140)], [(30, 40), (130, 140)])])
    def test_get_blocks_from_profile(self, profile1, features, expected):
        assert c.get_blocks_from_profile(features, profile1) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], False), ([0, 1], [1, 1], True),
                                                              ([0, 1, 1, 0], [-1, 1, 1, -2], False),
                                                              ([-1, -1, 1, 0], [-1, 1, 1, -2], True),
                                                              ([0, 1, -1], [-2, 1, 1], False)])
    def test_left_truncated(self, profile1, profile2, expected):
        assert c.left_truncated(profile1, profile2) == expected

    @pytest.mark.parametrize("profile1, profile2, expected", [([1], [1], False), ([1, 0], [1, 1], True),
                                                              ([0, 1, 1, 0], [-1, 1, 1, -2], False),
                                                              ([-1, 1, -1, 0], [-1, 1, 1, -2], True),
                                                              ([0, 1, -1], [-2, 1, 1], True)])
    def test_right_truncated(self, profile1, profile2, expected):
        assert c.right_truncated(profile1, profile2) == expected


class TestListToStr:
    def test_empty(self):
        assert "." == c.list_to_str([])

    @pytest.mark.parametrize("element_list, sep, expected", [([1], ",", "1"), ([1, 0, 2], ":", "1:0:2")])
    def test_not_empty(self, element_list, sep, expected):
        assert c.list_to_str(element_list, sep) == expected


class TestRangeOperations(unittest.TestCase):
    def test_overlaps(self):
        self.assertTrue(c.overlaps((100, 200), (150, 250)))
        self.assertFalse(c.overlaps((100, 200), (201, 300)))
        self.assertTrue(c.overlaps((100, 200), (200, 300)))

    def test_contains(self):
        self.assertTrue(c.contains((100, 300), (150, 250)))
        self.assertFalse(c.contains((150, 250), (100, 300)))
        self.assertTrue(c.contains((100, 300), (100, 300)))

    def test_left_of(self):
        self.assertTrue(c.left_of((100, 200), (201, 300)))
        self.assertFalse(c.left_of((100, 200), (200, 300)))

    def test_interval_len(self):
        self.assertEqual(c.interval_len((100, 200)), 101)
        self.assertEqual(c.interval_len((100, 100)), 1)

    def test_intervals_total_length(self):
        ranges = [(100, 200), (300, 400), (500, 600)]
        expected = 101 + 101 + 101
        self.assertEqual(c.intervals_total_length(ranges), expected)

    def test_max_range(self):
        range1 = (100, 200)
        range2 = (150, 250)
        result = c.max_range(range1, range2)
        self.assertEqual(result, (100, 250))

    def test_intersection_len(self):
        self.assertEqual(c.intersection_len((100, 200), (150, 250)), 51)
        self.assertEqual(c.intersection_len((100, 200), (300, 400)), 0)


class TestJunctions(unittest.TestCase):
    def test_junctions_from_blocks(self):
        blocks = [(100, 200), (300, 400), (500, 600)]
        junctions = c.junctions_from_blocks(blocks)
        expected = [(201, 299), (401, 499)]
        self.assertEqual(junctions, expected)

    def test_junctions_from_single_exon(self):
        blocks = [(100, 200)]
        junctions = c.junctions_from_blocks(blocks)
        self.assertEqual(junctions, [])

    def test_get_exons_from_junctions(self):
        region = (100, 600)
        introns = [(201, 299), (401, 499)]
        exons = c.get_exons(region, introns)
        expected = [(100, 200), (300, 400), (500, 600)]
        self.assertEqual(exons, expected)


class TestProfileFunctions(unittest.TestCase):
    def test_difference_in_present_features(self):
        profile1 = [1, 1, 0, 1, -1]
        profile2 = [1, -1, 0, 1, -1]
        diff = c.difference_in_present_features(profile1, profile2)
        self.assertEqual(diff, 1)  # Only position 1 differs

    def test_has_overlapping_features(self):
        profile1 = [1, 0, 1, 0]
        profile2 = [0, 1, 1, 0]
        self.assertTrue(c.has_overlapping_features(profile1, profile2))

        profile3 = [1, 0, 0, 0]
        profile4 = [0, 1, 1, 0]
        self.assertFalse(c.has_overlapping_features(profile3, profile4))


class TestUtilityFunctions(unittest.TestCase):
    def test_get_first_best_empty_list(self):
        result = c.get_first_best_from_sorted([])
        self.assertEqual(result, [])

    def test_rindex(self):
        lst = [1, 2, 3, 2, 1]
        self.assertEqual(c.rindex(lst, 2), 3)
        self.assertEqual(c.rindex(lst, 1), 4)

    def test_argmin(self):
        lst = [5, 2, 8, 1, 9]
        self.assertEqual(c.argmin(lst), 3)

    def test_reverse_complement(self):
        seq = "ATCG"
        self.assertEqual(c.reverse_complement(seq), "CGAT")
        seq = "AAAA"
        self.assertEqual(c.reverse_complement(seq), "TTTT")


############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

from enum import Enum, unique


@unique
class IsoQuantMode(Enum):
    bulk = 1
    tenX_v3 = 2
    tenX_v2 = 22
    curio = 3
    stereoseq_nosplit = 4
    stereoseq = 5
    visium_hd = 61
    visium_5prime = 7
    custom_sc = 10

    def needs_barcode_calling(self):
        return self in [IsoQuantMode.tenX_v3,
                        IsoQuantMode.tenX_v2,
                        IsoQuantMode.curio,
                        IsoQuantMode.stereoseq_nosplit,
                        IsoQuantMode.stereoseq,
                        IsoQuantMode.visium_hd,
                        IsoQuantMode.visium_5prime,
                        IsoQuantMode.custom_sc]

    def needs_pcr_deduplication(self):
        return self in [IsoQuantMode.tenX_v3,
                        IsoQuantMode.tenX_v2,
                        IsoQuantMode.curio,
                        IsoQuantMode.stereoseq_nosplit,
                        IsoQuantMode.stereoseq,
                        IsoQuantMode.visium_hd,
                        IsoQuantMode.visium_5prime,
                        IsoQuantMode.custom_sc]

    def produces_new_fasta(self):
        return self in [IsoQuantMode.stereoseq]

    def needs_barcode_iterator(self):
        return self in [IsoQuantMode.stereoseq_nosplit, IsoQuantMode.stereoseq]

    def enforces_single_thread(self):
        return False 


ISOQUANT_MODES = [x.name for x in IsoQuantMode]


############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################
import copy
import os
import sys
import logging
import gzip
import numpy
import math
from ssw import AlignmentMgr

from ..error_codes import IsoQuantExitCode


logger = logging.getLogger('IsoQuant')

# Try to import numba for JIT compilation of barcode conversion
try:
    from numba import njit, prange
    NUMBA_AVAILABLE = True
except (ImportError, SystemError):
    NUMBA_AVAILABLE = False


if NUMBA_AVAILABLE:
    @njit(cache=True, parallel=True)
    def _compute_2bit_numba(encoded, shifts, n_barcodes, seq_len):
        """Numba JIT-compiled 2-bit encoding computation with parallel execution."""
        result = numpy.zeros(n_barcodes, dtype=numpy.uint64)
        for i in prange(n_barcodes):
            val = numpy.uint64(0)
            for pos in range(seq_len):
                val |= numpy.uint64(encoded[i, pos]) << shifts[pos]
            result[i] = val
        return result


def find_polyt_start(seq, window_size = 16, polya_fraction = 0.75):
    polyA_count = int(window_size * polya_fraction)

    if len(seq) < window_size:
        return -1
    i = 0
    a_count = seq[0:window_size].count('T')
    while i < len(seq) - window_size:
        if a_count >= polyA_count:
            break
        first_base_a = seq[i] == 'T'
        new_base_a = i + window_size < len(seq) and seq[i + window_size] == 'T'
        if first_base_a and not new_base_a:
            a_count -= 1
        elif not first_base_a and new_base_a:
            a_count += 1
        i += 1

    if i >= len(seq) - window_size:
        return -1

    return i + max(0, seq[i:].find('TTTT'))


def find_polyt(seq, window_size = 16, polya_fraction = 0.75):
    polyA_count = int(window_size * polya_fraction)

    if len(seq) < window_size:
        return -1, -1
    i = 0
    a_count = seq[0:window_size].count('T')
    while i < len(seq) - window_size:
        if a_count >= polyA_count:
            break
        first_base_a = seq[i] == 'T'
        new_base_a = i + window_size < len(seq) and seq[i + window_size] == 'T'
        if first_base_a and not new_base_a:
            a_count -= 1
        elif not first_base_a and new_base_a:
            a_count += 1
        i += 1

    if i >= len(seq) - window_size:
        return -1, -1
    polyt_start = i + max(0, seq[i:].find('TTTT'))

    while i < len(seq) - window_size:
        if a_count < polyA_count:
            break
        first_base_a = seq[i] == 'T'
        new_base_a = i + window_size < len(seq) and seq[i + window_size] == 'T'
        if first_base_a and not new_base_a:
            a_count -= 1
        elif not first_base_a and new_base_a:
            a_count += 1
        i += 1

    if i >= len(seq) - window_size:
        polyt_end = len(seq) - 1
    else:
        last_t_pos = seq[i:i + window_size].rfind('TTTT')
        polyt_end = i + window_size - 1 if last_t_pos == -1 else i + last_t_pos + 4

    return polyt_start, polyt_end


base_comp = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', " ": " "}


def reverese_complement(my_seq):  ## obtain reverse complement of a sequence
    lms = list(map(lambda x: base_comp[x], my_seq))[::-1]
    return ''.join(lms)


def align_pattern_ssw(sequence, start, end, pattern, min_score=0):
    seq = sequence[start:end]
    align_mgr = AlignmentMgr(match_score=1, mismatch_penalty=1)
    align_mgr.set_read(pattern)
    align_mgr.set_reference(seq)
    alignment = align_mgr.align(gap_open=1, gap_extension=1)
    if alignment.optimal_score < min_score:
        return None, None, None, None, None
    return start + alignment.reference_start, start + alignment.reference_end, \
        alignment.read_start, alignment.read_end, alignment.optimal_score


def find_candidate_with_max_score_ssw(barcode_matches: list, read_sequence, min_score=10, score_diff=0, sufficient_score=0):
    best_match = [0, 0, 0]
    best_barcode = None
    second_best_score = 0

    align_mgr = AlignmentMgr(match_score=1, mismatch_penalty=1)
    align_mgr.set_reference(read_sequence)
    for barcode_match in barcode_matches:
        barcode = barcode_match[0]
        align_mgr.set_read(barcode)
        alignment = align_mgr.align(gap_open=1, gap_extension=1)
        if alignment.optimal_score < min_score:
            continue

        if alignment.optimal_score > best_match[0]:
            best_barcode = barcode
            second_best_score = best_match[0]
            best_match[0] = alignment.optimal_score
            best_match[1] = alignment.reference_start - alignment.read_start
            best_match[2] = alignment.reference_end + (len(barcode) - alignment.read_end)
        elif alignment.optimal_score  == best_match[0] and alignment.reference_start < best_match[1]:
            best_barcode = barcode
            second_best_score = best_match[0]
            best_match[1] = alignment.reference_start - alignment.read_start
            best_match[2] = alignment.reference_end + (len(barcode) - alignment.read_end)

        if alignment.optimal_score > sufficient_score > 0:
            # dirty hack to select first "sufficiently good" alignment
            break

    if best_match[0] - second_best_score < score_diff:
        return None, 0, 0, 0

    return best_barcode, best_match[0], best_match[1], best_match[2]


def find_candidate_with_max_score_ssw_var_len(barcode_matches: list, read_sequence, min_score=14, score_diff=1):
    best_match = [0, 0, 0, 0]
    second_best_match = [0, 0, 0, 0]
    best_barcode = None

    align_mgr = AlignmentMgr(match_score=1, mismatch_penalty=1)
    align_mgr.set_reference(read_sequence)
    for barcode_match in barcode_matches:
        barcode = barcode_match[0]
        align_mgr.set_read(barcode)
        alignment = align_mgr.align(gap_open=1, gap_extension=1)
        if alignment.optimal_score < min_score:
            continue

        ed = len(barcode) - alignment.optimal_score
        if alignment.optimal_score > best_match[0]:
            second_best_match = copy.copy(best_match)
            best_barcode = barcode
            best_match[0] = alignment.optimal_score
            best_match[1] = alignment.reference_start - alignment.read_start
            best_match[2] = alignment.reference_end + (len(barcode) - alignment.read_end)
            best_match[3] = ed
        elif alignment.optimal_score == best_match[0] and ed <= best_match[3]:
            second_best_match = copy.copy(best_match)
            best_barcode = barcode
            best_match[1] = alignment.reference_start - alignment.read_start
            best_match[2] = alignment.reference_end + (len(barcode) - alignment.read_end)
            best_match[3] = ed

    if best_barcode and best_match[0] < len(best_barcode) and best_match[0] - second_best_match[0] < score_diff:
        return None, 0, 0, 0

    return best_barcode, best_match[0], best_match[1], best_match[2]


def detect_exact_positions(sequence, start, end, kmer_size, pattern, pattern_occurrences: list,
                           min_score=0, start_delta=-1, end_delta=-1):
    pattern_index = None
    for i, p in enumerate(pattern_occurrences):
        if p[0] == pattern:
            pattern_index = i
            break
    if pattern_index is None:
        return None, None

    start_pos, end_pos, pattern_start, pattern_end, score  = None, None, None, None, 0
    last_potential_pos = -2*len(pattern)
    for match_position in pattern_occurrences[pattern_index][2]:
        if match_position - last_potential_pos < len(pattern):
            continue

        potential_start = start + match_position - len(pattern) + kmer_size
        potential_start = max(start, potential_start)
        potential_end = start + match_position + len(pattern) + 1
        potential_end = min(end, potential_end)
        alignment = \
            align_pattern_ssw(sequence, potential_start, potential_end, pattern, min_score)
        if alignment[4] is not None and alignment[4] > score:
            start_pos, end_pos, pattern_start, pattern_end, score = alignment

    if start_pos is None:
        return None, None

    if start_delta >= 0 and pattern_start > start_delta:
        return None, None
    if end_delta >= 0 and len(pattern) - pattern_end - 1 > end_delta:
        return None, None
    leftover_bases = len(pattern) - pattern_end - 1
    skipped_bases = pattern_start
    return max(0, start_pos - skipped_bases), min(end_pos + leftover_bases, len(sequence) - 1)


def detect_first_exact_positions(sequence, start, end, kmer_size, pattern, pattern_occurrences: list,
                           min_score=0, start_delta=-1, end_delta=-1):
    pattern_index = None
    for i, p in enumerate(pattern_occurrences):
        if p[0] == pattern:
            pattern_index = i
            break
    if pattern_index is None:
        return None, None

    start_pos, end_pos, pattern_start, pattern_end, score  = None, None, None, None, 0
    last_potential_pos = -2*len(pattern)
    for match_position in pattern_occurrences[pattern_index][2]:
        if match_position - last_potential_pos < len(pattern):
            continue

        potential_start = start + match_position - len(pattern) + kmer_size
        potential_start = max(start, potential_start)
        potential_end = start + match_position + len(pattern) + 1
        potential_end = min(end, potential_end)
        alignment = \
            align_pattern_ssw(sequence, potential_start, potential_end, pattern, min_score)
        if alignment[4] is not None:
            start_pos, end_pos, pattern_start, pattern_end, score = alignment
            break

    if start_pos is None:
        return None, None

    if start_delta >= 0 and pattern_start > start_delta:
        return None, None
    if end_delta >= 0 and len(pattern) - pattern_end - 1 > end_delta:
        return None, None
    leftover_bases = len(pattern) - pattern_end - 1
    skipped_bases = pattern_start
    return start_pos - skipped_bases, end_pos + leftover_bases


NUCL2BIN = {'A': 0, 'C': 1, 'G': 3, 'T': 2, 'a': 0, 'c': 1, 'g': 3, 't': 2}
BIN2NUCL = ["A", "C", "T", "G"]


def str_to_2bit(seq):
    kmer_idx = 0
    seq_len = len(seq)
    for i in range(seq_len):
        kmer_idx |= ((ord(seq[i]) & 6) >> 1) << ((seq_len - i - 1) * 2)
    return kmer_idx


def batch_str_to_2bit(barcodes, seq_len):
    """
    Convert a list/iterator of barcode strings to 2-bit encoded integers.

    Uses NumPy vectorization for ~50-100x speedup over individual str_to_2bit calls.
    For 500M barcodes, this reduces conversion time from hours to minutes.

    Args:
        barcodes: Iterable of barcode strings (all must be same length)
        seq_len: Length of each barcode (default 25 for Stereo-seq)

    Returns:
        numpy.ndarray of uint64 containing 2-bit encoded barcodes
    """
    # Convert iterator to list if needed, reading in chunks to show progress
    if not isinstance(barcodes, (list, numpy.ndarray)):
        logger.info("Loading barcodes from iterator...")
        barcode_list = []
        chunk_size = 10_000_000
        for bc in barcodes:
            barcode_list.append(bc)
            if len(barcode_list) % chunk_size == 0:
                logger.info("  Loaded %d barcodes..." % len(barcode_list))
        barcodes = barcode_list
        logger.info("  Loaded %d barcodes total" % len(barcodes))

    n_barcodes = len(barcodes)
    if n_barcodes == 0:
        return numpy.array([], dtype=numpy.uint64)

    logger.info("Converting %d barcodes to 2-bit encoding..." % n_barcodes)

    # Fast path: join all barcodes into single string, convert to bytes at once
    # This avoids Python loop overhead for encode()
    all_bytes = ''.join(barcodes).encode('ascii')
    byte_array = numpy.frombuffer(all_bytes, dtype=numpy.uint8).reshape(n_barcodes, seq_len)

    # Apply the encoding: (ord(char) & 6) >> 1
    # A=65 -> 0, C=67 -> 1, T=84 -> 2, G=71 -> 3
    encoded = (byte_array & 6) >> 1

    # Compute bit shifts for each position (highest position = leftmost)
    shifts = numpy.arange(seq_len - 1, -1, -1, dtype=numpy.uint64) * 2

    # Compute 2-bit encoding - use Numba if available for parallel execution
    if NUMBA_AVAILABLE:
        result = _compute_2bit_numba(encoded, shifts, n_barcodes, seq_len)
    else:
        # Fallback: vectorized NumPy loop
        result = numpy.zeros(n_barcodes, dtype=numpy.uint64)
        for pos in range(seq_len):
            result |= encoded[:, pos].astype(numpy.uint64) << shifts[pos]

    logger.info("Barcode conversion complete")
    return result


def batch_str_to_2bit_chunked(barcode_iterator, seq_len, chunk_size=50_000_000):
    """
    Convert barcodes to 2-bit encoding in memory-efficient chunks.

    Processes barcodes in chunks to avoid loading all strings into memory at once.
    Returns a single concatenated numpy array.

    Args:
        barcode_iterator: Iterator yielding barcode strings
        seq_len: Length of each barcode
        chunk_size: Number of barcodes to process per chunk

    Returns:
        numpy.ndarray of uint64 containing 2-bit encoded barcodes
    """
    all_results = []
    current_chunk = []
    total_processed = 0

    logger.info("Converting barcodes to 2-bit encoding (chunked mode)...")

    for bc in barcode_iterator:
        current_chunk.append(bc)
        if len(current_chunk) >= chunk_size:
            # Process this chunk
            chunk_result = _convert_chunk_to_2bit(current_chunk, seq_len)
            all_results.append(chunk_result)
            total_processed += len(current_chunk)
            logger.info("  Processed %d barcodes..." % total_processed)
            current_chunk = []

    # Process remaining barcodes
    if current_chunk:
        chunk_result = _convert_chunk_to_2bit(current_chunk, seq_len)
        all_results.append(chunk_result)
        total_processed += len(current_chunk)

    logger.info("Barcode conversion complete: %d barcodes" % total_processed)

    if not all_results:
        return numpy.array([], dtype=numpy.uint64)
    return numpy.concatenate(all_results)


def _convert_chunk_to_2bit(barcodes, seq_len):
    """Convert a chunk of barcodes to 2-bit encoding (internal helper)."""
    n_barcodes = len(barcodes)
    if n_barcodes == 0:
        return numpy.array([], dtype=numpy.uint64)

    # Fast path: join all barcodes into single string, convert to bytes at once
    all_bytes = ''.join(barcodes).encode('ascii')
    byte_array = numpy.frombuffer(all_bytes, dtype=numpy.uint8).reshape(n_barcodes, seq_len)

    # Apply encoding: (ord(char) & 6) >> 1
    encoded = (byte_array & 6) >> 1

    # Compute shifts
    shifts = numpy.arange(seq_len - 1, -1, -1, dtype=numpy.uint64) * 2

    # Compute result - use Numba if available for parallel execution
    if NUMBA_AVAILABLE:
        result = _compute_2bit_numba(encoded, shifts, n_barcodes, seq_len)
    else:
        result = numpy.zeros(n_barcodes, dtype=numpy.uint64)
        for pos in range(seq_len):
            result |= encoded[:, pos].astype(numpy.uint64) << shifts[pos]

    return result


def bit_to_str(seq, seq_len):
    # Convert numpy scalar to Python int for bitwise operations
    seq = int(seq)
    str_seq = ""
    for i in range(seq_len):
        str_seq += BIN2NUCL[(seq >> ((seq_len - i - 1) * 2)) & 3]
    return str_seq



def load_h5_barcodes_bit(h5_file_path, dataset_name='bpMatrix_1'):
    raise NotImplementedError()
    import h5py
    barcode_list = []
    with h5py.File(h5_file_path, 'r') as h5_file:
        dataset = numpy.array(h5_file[dataset_name])
        for row in dataset:
            for col in row:
                barcode_list.append(bit_to_str(int(col[0])))
    return barcode_list


def load_barcodes(inf, needs_iterator=False):
    if not os.path.isfile(inf):
        logger.critical("Barcode file '%s' does not exist.", inf)
        sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)

    if inf.endswith("h5") or inf.endswith("hdf5"):
        return load_h5_barcodes_bit(inf)

    if inf.endswith("gz") or inf.endswith("gzip"):
        handle = gzip.open(inf, "rt")
    else:
        handle = open(inf, "r")

    barcode_iterator = iter(l.strip().split()[0] for l in handle)
    if needs_iterator:
        return barcode_iterator

    return list(barcode_iterator)


def find_optimal_kmer_size(bc_len, bc_count, min_k=6, max_k=14, kmer_density=50):
    k = min_k
    while bc_count * (bc_len - k + 1)  * (bc_len - k + 1) > math.pow(4, k) * kmer_density:
        k += 1

    return min(k, max_k)



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Dict-based flexible detection result for universal molecule extraction.

Unlike the attribute-based result classes in base.py, ExtractionResult stores
detected elements in a dictionary, allowing flexible molecule structures.
Implements the BarcodeResult protocol.
"""

import logging
from collections import defaultdict
from typing import List

from .molecule_structure import MoleculeStructure, ElementType

logger = logging.getLogger('IsoQuant')


class DetectedElement:
    """Container for a single detected element's coordinates and sequence."""

    def __init__(self, start: int = -1, end: int = -1, score: int = -1, seq: str = None):
        self.start = start
        self.end = end
        self.score = score
        self.seq = seq


class ExtractionResult:
    """
    Dict-based flexible detection result for universal molecule extraction.

    Unlike the attribute-based result classes (BarcodeDetectionResult, etc.),
    ExtractionResult stores detected elements in a dictionary, allowing
    flexible molecule structures defined by MoleculeStructure.

    Implements the BarcodeResult protocol.

    Barcode/UMI elements are identified by naming convention:
    - Elements with prefix "barcode" (case-insensitive) are barcodes
    - Elements with prefix "umi" (case-insensitive) are UMIs
    """

    NOSEQ = "*"  # Sentinel for missing/undetected sequence (matches BarcodeDetectionResult.NOSEQ)

    def __init__(self, molecule_structure: MoleculeStructure, read_id: str, strand: str, detected_results: dict):
        """
        Initialize extraction result.

        Args:
            molecule_structure: Molecule structure definition
            read_id: Read identifier
            strand: Detected strand ('+', '-', or '.')
            detected_results: Dict mapping element names to DetectedElement objects
        """
        self.molecule_structure = molecule_structure
        self.read_id = read_id
        self.strand = strand
        self.detected_results = detected_results

    @property
    def _barcode_elements(self) -> List[str]:
        """Get barcode element names from molecule structure."""
        return self.molecule_structure.barcode_elements

    @property
    def _umi_elements(self) -> List[str]:
        """Get UMI element names from molecule structure."""
        return self.molecule_structure.umi_elements

    def get_barcode(self) -> str:
        """
        Get the detected barcode sequence.

        For molecule structures with multiple barcode elements (e.g., barcode1, barcode2),
        returns concatenated barcodes using '|' delimiter only when ALL parts are detected.
        Returns NOSEQ if any part is missing (partial barcodes are not usable).

        Returns:
            str: Concatenated barcode sequence or NOSEQ if any part is not detected.
        """
        parts = []
        for el_name in self._barcode_elements:
            if el_name in self.detected_results:
                detected = self.detected_results[el_name]
                if detected.seq and detected.seq != self.NOSEQ:
                    parts.append(detected.seq)
                    continue
            return self.NOSEQ
        return "|".join(parts) if parts else self.NOSEQ

    def get_umi(self) -> str:
        """
        Get the detected UMI sequence.

        For molecule structures with multiple UMI elements,
        returns all parts using '|' delimiter, with NOSEQ ('*') for undetected elements.

        Returns:
            str: Concatenated UMI sequence or NOSEQ if not detected.
        """
        if len(self._umi_elements) <= 1:
            for el_name in self._umi_elements:
                if el_name in self.detected_results:
                    detected = self.detected_results[el_name]
                    if detected.seq and detected.seq != self.NOSEQ:
                        return detected.seq
            return self.NOSEQ

        parts = []
        for el_name in self._umi_elements:
            if el_name in self.detected_results:
                detected = self.detected_results[el_name]
                if detected.seq and detected.seq != self.NOSEQ:
                    parts.append(detected.seq)
                    continue
            parts.append(self.NOSEQ)
        return "|".join(parts)

    def is_valid(self) -> bool:
        """Check if a valid barcode was detected (all parts must be present)."""
        return self.get_barcode() != self.NOSEQ

    def has_barcode(self) -> bool:
        """Check if a barcode was detected (alias for is_valid())."""
        return self.get_barcode() != self.NOSEQ

    def has_umi(self) -> bool:
        """Check if any valid UMI was detected."""
        return self.get_umi() != self.NOSEQ

    def get_barcode_score(self) -> int:
        """
        Get the best barcode alignment score.

        Returns the maximum score from all barcode elements, or -1 if none detected.
        """
        max_score = -1
        for el_name in self._barcode_elements:
            if el_name in self.detected_results:
                detected = self.detected_results[el_name]
                if detected.score > max_score:
                    max_score = detected.score
        return max_score

    def update_coordinates(self, delta: int) -> None:
        """
        Shift all genomic coordinates by delta.

        Used when processing read subsequences.

        Args:
            delta: Amount to shift coordinates
        """
        for detected in self.detected_results.values():
            if detected.start != -1:
                detected.start += delta
            if detected.end != -1:
                detected.end += delta

    def more_informative_than(self, other: 'ExtractionResult') -> bool:
        """
        Compare two results to determine which is more informative.

        Compares by number of detected elements, then total score.

        Args:
            other: Another detection result

        Returns:
            True if this result is more informative than other
        """
        self_count = sum(1 for d in self.detected_results.values() if d.start != -1)
        other_count = sum(1 for d in other.detected_results.values() if d.start != -1)
        if self_count != other_count:
            return self_count > other_count
        self_score = sum(d.score for d in self.detected_results.values() if d.score > 0)
        other_score = sum(d.score for d in other.detected_results.values() if d.score > 0)
        return self_score > other_score

    def get_additional_attributes(self) -> List[str]:
        """
        Get list of detected supplementary features (excluding barcode/UMI).

        Barcode and UMI elements are excluded because they are already
        tracked by ReadStats.bc_count and ReadStats.umi_count respectively.

        Returns:
            List of detected element names with " detected" suffix.
        """
        barcode_umi_set = set(self._barcode_elements + self._umi_elements)
        return ["%s detected" % el_name for el_name, d in self.detected_results.items()
                if d.start != -1 and el_name not in barcode_umi_set]

    def set_strand(self, strand: str) -> None:
        """Set the detected strand."""
        self.strand = strand

    def _get_supplementary_elements(self) -> List[str]:
        """Get list of element names that are not barcode, UMI, or concatenated/duplicated parts."""
        supplementary = []
        barcode_umi_set = set(self._barcode_elements + self._umi_elements)
        for el in self.molecule_structure:
            if el.element_type == ElementType.cDNA:
                continue
            if el.element_name in self.molecule_structure.elements_to_concatenate:
                continue
            if el.element_name in self.molecule_structure.duplicated_elements:
                continue
            if el.element_name not in barcode_umi_set:
                supplementary.append(el.element_name)
        return supplementary

    def __str__(self) -> str:
        """
        Format result as TSV line in standard format.

        Format: read_id, barcode, UMI, BC_score, valid_UMI, strand, [supplementary elements]
        This matches the format expected by the rest of the IsoQuant pipeline.
        """
        # Standard fields: read_id, barcode, UMI, BC_score, valid_UMI, strand
        res_str = "%s\t%s\t%s\t%d\t%s\t%s" % (
            self.read_id,
            self.get_barcode(),
            self.get_umi(),
            self.get_barcode_score(),
            self.has_umi(),
            self.strand
        )

        # Supplementary elements (non-barcode, non-UMI)
        barcode_umi_set = set(self._barcode_elements + self._umi_elements)
        for el in self.molecule_structure:
            if el.element_type == ElementType.cDNA:
                continue
            # Skip concatenated/duplicated parts (data lives under base name)
            if el.element_name in self.molecule_structure.elements_to_concatenate:
                continue
            if el.element_name in self.molecule_structure.duplicated_elements:
                continue
            # Skip barcode and UMI elements (already printed above)
            if el.element_name in barcode_umi_set:
                continue

            if el.element_type == ElementType.PolyT:
                if ElementType.PolyT.name not in self.detected_results:
                    res_str += "\t-1\t-1"
                    continue
                detected_element = self.detected_results[ElementType.PolyT.name]
                res_str += "\t%d\t%d" % (detected_element.start, detected_element.end)
            elif el.element_type.is_constant():
                if el.element_name not in self.detected_results:
                    res_str += "\t-1\t-1"
                    continue
                detected_element = self.detected_results[el.element_name]
                res_str += "\t%d\t%d" % (detected_element.start, detected_element.end)
            elif el.element_type.needs_sequence_extraction() or el.element_type.needs_correction():
                if el.element_name not in self.detected_results:
                    res_str += "\t-1\t-1\t%s" % self.NOSEQ
                    continue
                detected_element = self.detected_results[el.element_name]
                seq = detected_element.seq if detected_element.seq else self.NOSEQ
                res_str += "\t%d\t%d\t%s" % (detected_element.start, detected_element.end, seq)
        return res_str

    def header(self) -> str:
        """
        Get TSV header for result output in standard format.

        Format: #read_id, barcode, UMI, BC_score, valid_UMI, strand, [supplementary elements]
        """
        # Standard header fields
        header = "read_id\tbarcode\tUMI\tBC_score\tvalid_UMI\tstrand"

        # Supplementary elements (non-barcode, non-UMI)
        barcode_umi_set = set(self._barcode_elements + self._umi_elements)
        for el in self.molecule_structure:
            if el.element_type == ElementType.cDNA:
                continue
            # Skip concatenated/duplicated parts (data lives under base name)
            if el.element_name in self.molecule_structure.elements_to_concatenate:
                continue
            if el.element_name in self.molecule_structure.duplicated_elements:
                continue
            # Skip barcode and UMI elements (already in standard header)
            if el.element_name in barcode_umi_set:
                continue

            if el.element_type == ElementType.PolyT:
                header += "\tpolyT_start\tpolyT_end"
            elif el.element_type.needs_only_coordinates():
                header += "\t%s_start\t%s_end" % (el.element_name, el.element_name)
            elif el.element_type.needs_sequence_extraction() or el.element_type.needs_correction():
                header += "\t%s_start\t%s_end\t%s_seq" % (el.element_name, el.element_name, el.element_name)
        return header

    # TODO: add simple serialization


class ReadStats:
    """Statistics tracker for ExtractionResult objects."""

    def __init__(self):
        self.read_count = 0
        self.bc_count = 0
        self.umi_count = 0
        self.pattern_counts = defaultdict(int)

    def add_read(self, barcode_detection_result: ExtractionResult) -> None:
        """Add an ExtractionResult to statistics."""
        self.read_count += 1
        # Count valid barcodes and UMIs
        if barcode_detection_result.has_barcode():
            self.bc_count += 1
        if barcode_detection_result.has_umi():
            self.umi_count += 1
        # Count individual detected elements
        for el in barcode_detection_result.detected_results:
            if barcode_detection_result.detected_results[el].start != -1:
                self.pattern_counts[el] += 1

    def __str__(self) -> str:
        human_readable_str = "Total reads:\t%d\n" % self.read_count
        human_readable_str += "Barcode detected:\t%d\n" % self.bc_count
        human_readable_str += "UMI detected:\t%d\n" % self.umi_count
        for a in self.pattern_counts:
            human_readable_str += "%s:\t%d\n" % (a, self.pattern_counts[a])
        return human_readable_str
